"""
Composio server object collections
"""

import base64
import difflib
import json
import os
import time
import traceback
import typing as t
import warnings
from concurrent.futures import Future, ThreadPoolExecutor
from unittest import mock

import pysher
import requests
import typing_extensions as te
from pydantic import BaseModel, ConfigDict, Field
from pysher.channel import Channel as PusherChannel
from pysher.connection import Connection as PusherConnection

from composio.client.base import Collection
from composio.client.endpoints import v1, v2
from composio.client.enums import (
    Action,
    ActionType,
    App,
    AppType,
    Tag,
    TagType,
    Trigger,
    TriggerType,
)
from composio.client.exceptions import ComposioClientError, ComposioSDKError
from composio.constants import PUSHER_CLUSTER, PUSHER_KEY
from composio.utils import logging


if t.TYPE_CHECKING:
    from composio.client import Composio

AUTH_SCHEMES = ("OAUTH2", "OAUTH1", "API_KEY", "BASIC", "BEARER_TOKEN")
AuthSchemeType = t.Literal["OAUTH2", "OAUTH1", "API_KEY", "BASIC", "BEARER_TOKEN"]


def to_trigger_names(
    triggers: t.Union[t.List[str], t.List[Trigger], t.List[TriggerType]],
) -> str:
    """Get trigger names as a string."""
    return ",".join([Trigger(trigger).slug for trigger in triggers])


class AuthConnectionParamsModel(BaseModel):
    """
    Authentication connection parameters.
    """

    scope: t.Optional[str] = None
    base_url: t.Optional[str] = None
    client_id: t.Optional[str] = None
    token_type: t.Optional[str] = None
    access_token: t.Optional[str] = None
    client_secret: t.Optional[str] = None
    consumer_id: t.Optional[str] = None
    consumer_secret: t.Optional[str] = None
    headers: t.Optional[dict] = None
    queryParams: t.Optional[dict] = None


class ConnectedAccountModel(BaseModel):
    """
    Connected account data model.
    """

    id: str
    status: str
    createdAt: str
    updatedAt: str
    appUniqueId: str
    appName: str
    integrationId: str
    connectionParams: AuthConnectionParamsModel

    clientUniqueUserId: t.Optional[str] = None
    entityId: t.Optional[str] = None

    # Override arbitrary model config.
    model_config: ConfigDict = ConfigDict(  # type: ignore
        arbitrary_types_allowed=True,
    )


class ConnectionRequestModel(BaseModel):
    """Connection request model."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    connectionStatus: str
    connectedAccountId: str
    redirectUrl: t.Optional[str] = None

    def save_user_access_data(
        self,
        client: "Composio",
        field_inputs: t.Dict,
        redirect_url: t.Optional[str] = None,
        entity_id: t.Optional[str] = None,
    ) -> t.Dict:
        """Save user access data."""
        connected_account = client.connected_accounts.get(  # type: ignore
            connection_id=self.connectedAccountId,
        )
        resp = client.http.post(
            url=str(v1 / "connectedAccounts"),
            json={
                "integrationId": connected_account.integrationId,
                "data": field_inputs,
                "redirectUri": redirect_url,
                "userUuid": entity_id,
            },
        )
        return resp.json()

    def wait_until_active(
        self,
        client: "Composio",
        timeout=60,
    ) -> "ConnectedAccountModel":
        start_time = time.time()
        while time.time() - start_time < timeout:
            connection = client.connected_accounts.get(  # type: ignore
                connection_id=self.connectedAccountId,
            )
            if connection.status == "ACTIVE":
                return connection
            time.sleep(1)

        # TODO: Replace with timeout error.
        raise ComposioClientError(
            "Connection did not become active within the timeout period."
        )


class ConnectionAuthParam(BaseModel):
    in_: str = Field(alias="in")
    name: str
    value: str


class ConnectionParams(BaseModel):
    body: t.Dict
    base_url: str
    parameters: t.List[ConnectionAuthParam]


class ConnectedAccounts(Collection[ConnectedAccountModel]):
    """Collection of connected accounts."""

    model = ConnectedAccountModel
    endpoint = v1 / "connectedAccounts"

    @t.overload  # type: ignore
    def get(self) -> t.List[ConnectedAccountModel]:
        """
        Get all connected accounts

        :return: List of Connected accounts
        """

    @t.overload  # type: ignore
    def get(self, connection_id: str) -> ConnectedAccountModel:
        """
        Get an account by connection ID

        :param connection_id: ID of the connection to filter by
        :return: Connected account
        """

    @t.overload
    def get(
        self,
        connection_id: t.Optional[str] = None,
        entity_ids: t.Optional[t.Sequence[str]] = None,
        active: bool = False,
    ) -> t.List[ConnectedAccountModel]:
        """
        Get a list of connected accounts by entity IDs

        :param entity_ids: List of entity IDs to filter by
        :param active: Returns account which are currently active
        :return: List of connected accounts
        """

    def get(
        self,
        connection_id: t.Optional[str] = None,
        entity_ids: t.Optional[t.Sequence[str]] = None,
        active: bool = False,
    ) -> t.Union[ConnectedAccountModel, t.List[ConnectedAccountModel]]:
        """
        Get a list of connected accounts.

        :param entity_ids: List of entity IDs to filter by
        :param connection_id: Return the connected account by a specific
                connection ID
        :param active: Returns account which are currently active
        :return: List of connected accounts
        """
        entity_ids = entity_ids or ()
        if connection_id is not None and len(entity_ids) > 0:
            raise ComposioClientError(
                message="Cannot use both `connection_id` and `entity_ids` parameters as filter"
            )

        if connection_id is not None:
            response = self._raise_if_required(
                self.client.http.get(
                    url=str(self.endpoint / connection_id),
                )
            )
            return self.model(**response.json())

        queries = {}
        if len(entity_ids) > 0:
            queries["user_uuid"] = ",".join(entity_ids)

        if active:
            queries["showActiveOnly"] = "true"

        response = self._raise_if_required(
            self.client.http.get(
                url=str(self.endpoint(queries=queries)),
            )
        )
        return [self.model(**account) for account in response.json().get("items", [])]

    def initiate(
        self,
        integration_id: str,
        entity_id: t.Optional[str] = None,
        params: t.Optional[t.Dict] = None,
        labels: t.Optional[t.List] = None,
        redirect_url: t.Optional[str] = None,
    ) -> ConnectionRequestModel:
        """Initiate a new connected account."""
        response = self._raise_if_required(
            response=self.client.http.post(
                url=str(self.endpoint),
                json={
                    "integrationId": integration_id,
                    "userUuid": entity_id,
                    "data": params or {},
                    "labels": labels or [],
                    "redirectUri": redirect_url,
                },
            )
        )
        return ConnectionRequestModel(**response.json())

    def info(self, connection_id: str) -> ConnectionParams:
        response = self._raise_if_required(
            self.client.http.get(
                url=str(self.endpoint / connection_id / "info"),
            )
        )
        return ConnectionParams(**response.json())


class AuthSchemeField(BaseModel):
    """Auth scheme field."""

    name: str
    display_name: t.Optional[str] = None
    description: str

    type: str
    default: t.Optional[str] = None
    required: bool = False
    expected_from_customer: bool = True

    get_current_user_endpoint: t.Optional[str] = None


class AppAuthScheme(BaseModel):
    """App authenticatio scheme."""

    scheme_name: str
    auth_mode: AuthSchemeType
    fields: t.List[AuthSchemeField]

    proxy: t.Optional[t.Dict] = None
    authorization_url: t.Optional[str] = None
    token_url: t.Optional[str] = None
    default_scopes: t.Optional[t.List] = None
    token_response_metadata: t.Optional[t.List] = None
    client_id: t.Optional[str] = None
    client_secret: t.Optional[str] = None


class AppModel(BaseModel):
    """App data model."""

    name: str
    key: str
    appId: str
    description: str
    categories: t.List[str]
    meta: t.Dict

    logo: t.Optional[str] = None
    docs: t.Optional[str] = None
    group: t.Optional[str] = None
    status: t.Optional[str] = None
    enabled: bool = False
    no_auth: bool = False
    auth_schemes: t.Optional[t.List[AppAuthScheme]] = None
    testConnectors: t.Optional[t.List[t.Dict[str, t.Any]]] = None
    documentation_doc_text: t.Optional[str] = None
    configuration_docs_text: t.Optional[str] = None


class Apps(Collection[AppModel]):
    """Collection of composio apps.."""

    model = AppModel
    endpoint = v1.apps

    @t.overload  # type: ignore
    def get(self) -> t.List[AppModel]:
        """Get available apps."""

    @t.overload
    def get(self, name: t.Optional[str] = None) -> AppModel:
        """Get a specific app."""

    def get(self, name: t.Optional[str] = None) -> t.Union[AppModel, t.List[AppModel]]:
        """Get apps."""
        if name is not None:
            return self.model(
                **self._raise_if_required(
                    response=self.client.http.get(
                        url=str(self.endpoint / name),
                    )
                ).json()
            )

        return super().get(queries={})


class TypeModel(BaseModel):
    type: str


class TriggerPayloadPropertyModel(BaseModel):
    """Trigger payload property data model."""

    description: str
    title: str
    type: t.Optional[str] = None
    anyOf: t.Optional[t.List[TypeModel]] = None

    examples: t.Optional[t.List] = None


class TriggerPayloadModel(BaseModel):
    """Trigger payload data model."""

    properties: t.Dict[str, TriggerPayloadPropertyModel]
    title: str
    type: t.Optional[str] = None
    anyOf: t.Optional[t.List[TypeModel]] = None

    required: t.Optional[t.List[str]] = None


class TriggerConfigPropertyModel(BaseModel):
    """Trigger config property data model."""

    description: str
    title: str

    type: t.Optional[str] = None


class TriggerConfigModel(BaseModel):
    """Trigger config data model."""

    properties: t.Dict[str, TriggerConfigPropertyModel]
    title: str

    type: t.Optional[str] = None
    required: t.Optional[t.List[str]] = None


class CallbackModel(dict):
    """Trigger callback model."""


class CallbackCollection(Collection[CallbackModel]):
    """Callback collection for triggers."""

    model = CallbackModel
    endpoint = v1.triggers

    def set(self, url: str) -> CallbackModel:
        """Set callback URL."""
        response = self._raise_if_required(
            response=self.client.http.post(
                url=str(self.endpoint / "setCallbackURL"),
                json={
                    "callbackURL": url,
                },
            )
        )
        return response.json()

    def get(self) -> str:  # type: ignore
        """Get current callback URL."""
        response = self._raise_if_required(
            response=self.client.http.get(
                url=str(self.endpoint / "callback_url"),
            )
        )
        return response.json().get("callbackURL")


class TriggerModel(BaseModel):
    """Trigger data model."""

    name: str
    display_name: str
    description: str
    payload: TriggerPayloadModel
    config: TriggerConfigModel
    instructions: str
    appId: str
    appKey: str
    appName: str
    count: int
    enabled: bool

    logo: t.Optional[str] = None


class SuccessExecuteActionResponseModel(BaseModel):
    """Success execute action response data model."""

    successfull: bool
    data: t.Dict
    error: t.Optional[str] = None


class FileType(BaseModel):
    name: str = Field(
        ..., description="File name, contains extension to indetify the file type"
    )
    content: str = Field(..., description="File content in base64")


class Connection(BaseModel):
    id: str
    integrationId: str
    clientUniqueUserId: str
    status: str


class Metadata(BaseModel):
    id: str
    connectionId: str
    triggerName: str
    triggerData: str
    triggerConfig: t.Dict[str, t.Any]
    connection: Connection


class TriggerEventData(BaseModel):
    """Trigger event payload."""

    appName: str
    payload: dict
    originalPayload: t.Dict[str, t.Any]
    metadata: Metadata

    clientId: t.Optional[int] = None


class _ChunkedTriggerEventData(BaseModel):
    """Cunked trigger event data model."""

    id: str
    index: int
    chunk: str
    final: bool


class _TriggerEventFilters(te.TypedDict):
    """Trigger event filterset."""

    app_name: te.NotRequired[AppType]
    trigger_id: te.NotRequired[str]
    connection_id: te.NotRequired[str]
    trigger_name: te.NotRequired[TriggerType]
    entity_id: te.NotRequired[str]
    integration_id: te.NotRequired[str]


TriggerCallback = t.Callable[[TriggerEventData], None]


class TriggerSubscription(logging.WithLogger):
    """Trigger subscription."""

    _pusher: pysher.Pusher
    _channel: PusherChannel
    _connection: PusherConnection
    _alive: bool

    def __init__(self, client: "Composio") -> None:
        """Initialize subscription object."""
        logging.WithLogger.__init__(self)
        self.client = client
        self._alive = False
        self._chunks: t.Dict[str, t.Dict[int, str]] = {}
        self._callbacks: t.List[t.Tuple[TriggerCallback, _TriggerEventFilters]] = []

    def validate_filters(self, filters: _TriggerEventFilters):
        docs_link_msg = "\nRead more here: https://docs.composio.dev/introduction/intro/quickstart_3"
        if not isinstance(filters, dict):
            raise ComposioSDKError(
                "Expected filters to be a dictionary" + docs_link_msg
            )

        expected_filters = list(_TriggerEventFilters.__annotations__)
        for filter, value in filters.items():
            if filter not in expected_filters:
                error_msg = f"Unexpected filter {filter!r}"
                possible_values = difflib.get_close_matches(
                    filter, expected_filters, n=1
                )
                if possible_values:
                    (possible_value,) = possible_values
                    error_msg += f" Did you mean {possible_value!r}?"
                raise ComposioSDKError(error_msg + docs_link_msg)

            # Validate app name
            if filter == "app_name":
                if isinstance(value, App):
                    value = value.slug

                elif not isinstance(value, str):
                    raise ComposioSDKError(
                        f"Expected 'app_name' to be App or str, found {value!r}"
                        + docs_link_msg
                    )

                # Our enums are in uppercase but we accept lowercase ones too.
                value = value.upper()

                # Ensure the app exists
                app_names = list(App.iter())
                if value not in app_names:
                    error_msg = f"App {value!r} does not exist."
                    possible_values = difflib.get_close_matches(value, app_names, n=1)
                    if possible_values:
                        (possible_value,) = possible_values
                        error_msg += f" Did you mean {possible_value!r}?"

                    raise ComposioSDKError(error_msg + docs_link_msg)

                # Ensure at least one of the app's triggers are enabled on the account.
                active_triggers = [
                    trigger.triggerName for trigger in self.client.active_triggers.get()
                ]
                apps_for_triggers = {
                    Trigger(trigger).app.upper() for trigger in active_triggers
                }
                if value not in apps_for_triggers:
                    error_msg = (
                        f"App {value!r} has no triggers enabled on your account.\n"
                        "Find the possible triggers by running `composio triggers`."
                    )
                    raise ComposioSDKError(error_msg + docs_link_msg)

            # Validate trigger name
            if filter == "trigger_name":
                if isinstance(value, Trigger):
                    value = value.slug
                elif not isinstance(value, str):
                    raise ComposioSDKError(
                        f"Expected 'trigger_name' to be Trigger or str, found {value!r}"
                        + docs_link_msg
                    )

                # Our enums are in uppercase but we accept lowercase ones too.
                value = value.upper()

                # Ensure the trigger exists
                trigger_names = list(Trigger.iter())
                if value not in trigger_names:
                    error_msg = f"Trigger {value!r} does not exist."
                    possible_values = difflib.get_close_matches(
                        value, trigger_names, n=1
                    )
                    if possible_values:
                        (possible_value,) = possible_values
                        error_msg += f" Did you mean {possible_value!r}?"

                    raise ComposioSDKError(error_msg + docs_link_msg)

                # Ensure the trigger is added on your account
                active_triggers = [
                    trigger.triggerName for trigger in self.client.active_triggers.get()
                ]
                if value not in active_triggers:
                    error_msg = (
                        f"Trigger {value!r} is not enabled on your account.\nEnable"
                        f" the trigger by doing `composio triggers enable {value}`."
                    )
                    raise ComposioSDKError(error_msg + docs_link_msg)

    def callback(
        self,
        filters: t.Optional[_TriggerEventFilters] = None,
    ) -> t.Callable[[TriggerCallback], TriggerCallback]:
        """Register a trigger callaback."""
        # Ensure filters is the right type before we stuff it in the callbacks
        if filters is not None:
            self.validate_filters(filters)

        def _wrap(f: TriggerCallback) -> TriggerCallback:
            self._callbacks.append((f, filters or {}))
            return f

        return _wrap

    def _handle_callback(
        self,
        callback: TriggerCallback,
        data: TriggerEventData,
        filters: _TriggerEventFilters,
    ) -> t.Any:
        """Handle callback."""
        for name, check in (
            ("app_name", data.appName),
            ("trigger_id", data.metadata.id),
            ("connection_id", data.metadata.connectionId),
            ("trigger_name", data.metadata.triggerName),
            ("entity_id", data.metadata.connection.clientUniqueUserId),
            ("integration_id", data.metadata.connection.integrationId),
        ):
            value = filters.get(name)
            if value is None or str(value).lower() == check.lower():
                continue

            self.logger.debug(
                f"Skipping `{callback.__name__}` since "
                f"`{name}` filter does not match the event metadata",
            )
            return None

        try:
            return callback(data)
        except BaseException:
            self.logger.info(
                f"Error executing `{callback.__name__}` for "
                f"event `{data.metadata.triggerName}` "
                f"with error:\n {traceback.format_exc()}"
            )
            return None

    def _parse_payload(self, event: str) -> t.Optional[TriggerEventData]:
        """Parse event payload."""
        try:
            return TriggerEventData(**json.loads(event))
        except Exception as e:
            self.logger.warning(f"Error decoding payload: {e}")
            return None

    def handle_event(self, event: str) -> None:
        """Filter events and call the callback function."""
        data = self._parse_payload(event=event)
        if data is None:
            self.logger.error(f"Error parsing trigger payload: {event}")
            return

        self.logger.info(
            f"Received trigger event with trigger ID: {data.metadata.id} "
            f"and trigger name: {data.metadata.triggerName}"
        )
        awaitables: t.List[Future] = []
        with ThreadPoolExecutor() as executor:
            for callback, filters in self._callbacks:
                awaitables.append(
                    executor.submit(
                        self._handle_callback,
                        callback,
                        data,
                        filters,
                    )
                )
        _ = [future.result() for future in awaitables]

    def handle_chunked_events(self, event: str) -> None:
        """Handle chunked events."""
        data = _ChunkedTriggerEventData(**json.loads(event))
        if data.id not in self._chunks:
            self._chunks[data.id] = {}

        self._chunks[data.id][data.index] = data.chunk
        if data.final:
            _chunks = self._chunks.pop(data.id)
            self.handle_event(
                event="".join([_chunks[idx] for idx in sorted(_chunks)]),
            )

    def is_alive(self) -> bool:
        """Check if subscription is live."""
        return self._alive

    def has_errored(self) -> bool:
        """Check if the connection errored and disconnected."""
        return self._connection.socket is None or self._connection.socket.has_errored

    def set_alive(self) -> None:
        """Set `_alive` to True."""
        self._alive = True

    @te.deprecated("Use `wait_forever` instead")
    def listen(self) -> None:
        """Wait infinitely."""
        self.wait_forever()

    def wait_forever(self) -> None:
        """Wait infinitely."""
        while self.is_alive() and not self.has_errored():
            time.sleep(1)

    def stop(self) -> None:
        """Stop the trigger listener."""
        self._connection.disconnect()
        self._alive = False

    def restart(self) -> None:
        """Restart the subscription connection"""
        self._connection.disconnect()
        self._connection._connect()  # pylint: disable=protected-access


class _PusherClient(logging.WithLogger):
    """Pusher client for Composio SDK."""

    def __init__(self, client_id: str, client: "Composio") -> None:
        """Initialize pusher client."""
        super().__init__()
        self.client_id = client_id
        self.client = client
        self.api_key = self.client.api_key
        self.base_url = self.client.http.base_url
        self.subscription = TriggerSubscription(client=self.client)

    def _get_connection_handler(
        self,
        client_id: str,
        pusher: pysher.Pusher,
        subscription: TriggerSubscription,
    ) -> t.Callable[[str], None]:
        def _connection_handler(_: str) -> None:
            channel = t.cast(
                PusherChannel,
                pusher.subscribe(
                    channel_name=f"private-{client_id}_triggers",
                ),
            )
            channel.bind(
                event_name="trigger_to_client",
                callback=subscription.handle_event,
            )
            channel.bind(
                event_name="chunked-trigger_to_client",
                callback=subscription.handle_chunked_events,
            )
            subscription.set_alive()
            subscription._channel = channel  # pylint: disable=protected-access
            subscription._connection = (  # pylint: disable=protected-access
                channel.connection
            )

        return _connection_handler

    def connect(self, timeout: float = 15.0) -> TriggerSubscription:
        """Connect to Pusher channel for given client ID."""
        # Make a request to the Pusher webhook endpoint
        try:
            response = requests.post(
                url=f"{self.base_url}/v1/triggers/pusher",
                json={
                    "time": int(time.time() * 1000),  # Current time in milliseconds
                    "events": [
                        {
                            "name": "channel_occupied",
                            "channel": f"private-{self.client_id}_triggers",
                        }
                    ],
                },
                timeout=timeout,
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()
        except requests.RequestException as e:
            self.logger.error(f"Failed to send Pusher webhook: {e}")

        pusher = pysher.Pusher(
            key=PUSHER_KEY,
            cluster=PUSHER_CLUSTER,
            auth_endpoint=f"{self.base_url}/v1/client/auth/pusher_auth?fromPython=true",
            auth_endpoint_headers={
                "x-api-key": self.api_key,
            },
        )

        # Patch pusher logger
        pusher.connection.logger = mock.MagicMock()  # type: ignore
        pusher.connection.bind(
            "pusher:connection_established",
            self._get_connection_handler(
                client_id=self.client_id,
                pusher=pusher,
                subscription=self.subscription,
            ),
        )
        pusher.connect()

        # Wait for connection to get established
        deadline = time.time() + timeout
        while time.time() < deadline:
            if self.subscription.is_alive():
                self.subscription._pusher = pusher  # pylint: disable=protected-access
                return self.subscription
            time.sleep(0.5)

        raise TimeoutError(
            "Timed out while waiting for trigger listener to be established"
        )


class Triggers(Collection[TriggerModel]):
    """Collection of triggers."""

    model = TriggerModel
    endpoint = v1.triggers
    callbacks: CallbackCollection

    def __init__(self, client: "Composio") -> None:
        """Initialize triggers collections."""
        super().__init__(client)
        self.callbacks = CallbackCollection(
            client=self.client,
        )

    def get(  # type: ignore
        self,
        trigger_names: t.Optional[t.List[TriggerType]] = None,
        apps: t.Optional[t.List[str]] = None,
    ) -> t.List[TriggerModel]:
        """
        List active triggers

        :param trigger_names: Trigger names to filter by, can be a list of strings or Trigger objects
        :param app_names: App names to filter by
        :return: List of triggers filtered by provided parameters
        """
        queries = {}
        if trigger_names is not None and len(trigger_names) > 0:
            queries["triggerIds"] = to_trigger_names(trigger_names)
        if apps is not None and len(apps) > 0:
            queries["appNames"] = ",".join(apps)
        return super().get(queries=queries)

    def enable(
        self, name: str, connected_account_id: str, config: t.Dict[str, t.Any]
    ) -> t.Dict:
        """
        Enable a trigger

        :param name: Name of the trigger
        :param connected_account_id: ID of the relevant connected account
        """
        response = self._raise_if_required(
            self.client.http.post(
                url=str(self.endpoint.enable / connected_account_id / name),
                json={"triggerConfig": config},
            )
        )
        return response.json()

    def disable(self, id: str) -> t.Dict:
        """
        Disable a trigger

        :param name: Name of the trigger
        :param connected_account_id: ID of the relevant connected account
        """
        response = self._raise_if_required(
            self.client.http.patch(
                url=str(self.endpoint / "instance" / id / "status"),
                json={
                    "enabled": False,
                },
            )
        )
        return response.json()

    def subscribe(self, timeout: float = 15.0) -> TriggerSubscription:
        """Subscribe to a trigger and receive trigger events."""
        self.logger.info("Creating trigger subscription")
        response = self._raise_if_required(
            response=self.client.http.get(
                url="/v1/client/auth/client_info",
            )
        )
        client_id = response.json().get("client", {}).get("id")
        if client_id is None:
            raise ComposioClientError("Error fetching client ID")

        pusher = _PusherClient(
            client_id=client_id,
            client=self.client,
        )
        return pusher.connect(
            timeout=timeout,
        )


class ActiveTriggerModel(BaseModel):
    """Active trigger data model."""

    id: str
    connectionId: str
    triggerName: str
    triggerConfig: dict


class ActiveTriggers(Collection[ActiveTriggerModel]):
    """Collection of active triggers."""

    model = ActiveTriggerModel
    endpoint = v1.triggers / "active_triggers"

    _list_key = "triggers"

    def get(  # type: ignore
        self,
        trigger_ids: t.Optional[t.List[str]] = None,
        connected_account_ids: t.Optional[t.List[str]] = None,
        integration_ids: t.Optional[t.List[str]] = None,
        trigger_names: t.Optional[t.List[t.Union[str, Trigger]]] = None,
    ) -> t.List[ActiveTriggerModel]:
        """List active triggers."""
        trigger_ids = trigger_ids or []
        connected_account_ids = connected_account_ids or []
        integration_ids = integration_ids or []
        trigger_names = trigger_names or []
        queries = {}
        if len(trigger_ids) > 0:
            queries["triggerIds"] = ",".join(trigger_ids)
        if len(connected_account_ids) > 0:
            queries["connectedAccountIds"] = ",".join(connected_account_ids)
        if len(integration_ids) > 0:
            queries["integrationIds"] = ",".join(integration_ids)
        if len(trigger_names) > 0:
            queries["triggerNames"] = to_trigger_names(trigger_names)
        return self._raise_if_empty(super().get(queries=queries))


def _check_file_uploadable(param_field: dict) -> bool:
    return (
        isinstance(param_field, dict)
        and (param_field.get("title") in ["File", "FileType"])
        and all(
            field_name in param_field.get("properties", {})
            for field_name in ["name", "content"]
        )
    )


def _check_file_downloadable(param_field: dict) -> bool:
    return set(param_field.keys()) == {"name", "content"}


class ActionParametersModel(BaseModel):
    """Action parameter data models."""

    properties: t.Dict[str, t.Any]
    title: str
    type: str

    required: t.Optional[t.List[str]] = None


class ActionResponseModel(BaseModel):
    """Action response data model."""

    properties: t.Dict[str, t.Any]
    title: str
    type: str

    required: t.Optional[t.List[str]] = None


class ActionModel(BaseModel):
    """Action data model."""

    name: str
    display_name: t.Optional[str] = None
    parameters: ActionParametersModel
    response: ActionResponseModel
    appName: str
    appId: str
    tags: t.List[str]
    enabled: bool = False

    logo: t.Optional[str] = None
    description: t.Optional[str] = None


ParamPlacement = t.Literal["header", "path", "query", "subdomain"]


class CustomAuthParameter(te.TypedDict):
    in_: ParamPlacement
    name: str
    value: str


class CustomAuthObject(BaseModel):
    body: t.Dict = Field(default_factory=lambda: {})
    base_url: t.Optional[str] = None
    parameters: t.List[CustomAuthParameter] = Field(default_factory=lambda: [])


class Actions(Collection[ActionModel]):
    """Collection of composio actions.."""

    model = ActionModel
    endpoint = v2.actions

    # TODO: Overload
    def get(  # type: ignore
        self,
        actions: t.Optional[t.Sequence[ActionType]] = None,
        apps: t.Optional[t.Sequence[AppType]] = None,
        tags: t.Optional[t.Sequence[TagType]] = None,
        limit: t.Optional[int] = None,
        use_case: t.Optional[str] = None,
        allow_all: bool = False,
    ) -> t.List[ActionModel]:
        """
        Get a list of apps by the specified filters.

        :param actions: Filter by the list of Actions.
        :param apps: Filter by the list of Apps.
        :param tags: Filter by the list of given Tags.
        :param limit: Limit the number of actions to a specific number.
        :param use_case: Filter by use case.
        :param allow_all: Allow querying all of the actions for a specific
                        app
        :return: List of actions
        """
        actions = t.cast(t.List[Action], [Action(action) for action in actions or []])
        apps = t.cast(t.List[App], [App(app) for app in apps or []])
        tags = t.cast(t.List[Tag], [Tag(tag) for tag in tags or []])

        # Filter out local apps and actions
        local_apps = [app for app in apps if app.is_local]
        local_actions = [action for action in actions if action.is_local]
        apps = [app for app in apps if not app.is_local]
        actions = [action for action in actions if not action.is_local]
        only_local_apps = (
            len(apps) == 0
            and len(actions) == 0
            and (len(local_apps) > 0 or len(local_actions) > 0)
        )
        if only_local_apps:
            local_items = self.client.local.get_action_schemas(
                apps=local_apps,
                actions=local_actions,
                tags=tags,
            )
            return [self.model(**item) for item in local_items]

        if len(actions) > 0 and len(apps) > 0:
            raise ComposioClientError(
                "Error retrieving Actions, Both actions and apps "
                "cannot be used as filters at the same time."
            )

        if len(actions) > 0 and len(tags) > 0:
            raise ComposioClientError(
                "Error retrieving Actions, Both actions and tags "
                "cannot be used as filters at the same time."
            )

        if len(apps) > 0 and len(tags) == 0 and not allow_all:
            warnings.warn(
                "Using all the actions of an app is not recommended. "
                "Please use tags to filter actions or provide specific actions. "
                "We just pass the important actions to the agent, but this is not meant "
                "to be used in production. Check out https://docs.composio.dev/sdk/python/actions for more information.",
                UserWarning,
            )
            tags = ["important"]

        if (
            len(actions) == 0
            and len(apps) == 0
            and len(tags) == 0
            and allow_all
            and len(local_apps) == 0
            and len(local_actions) == 0
        ):
            response = self._raise_if_required(
                response=self.client.http.get(
                    url=str(self.endpoint),
                )
            )
            return [self.model(**action) for action in response.json().get("items")]

        queries: t.Dict[str, str] = {}
        if use_case is not None and use_case != "":
            queries["useCase"] = use_case

        if len(apps) > 0:
            queries["apps"] = ",".join(list(map(lambda x: t.cast(App, x).slug, apps)))

        if len(actions) > 0:
            queries["apps"] = ",".join(
                set(map(lambda x: t.cast(Action, x).app, actions))
            )

        if limit is not None:
            queries["limit"] = str(limit)

        response = self._raise_if_required(
            response=self.client.http.get(
                url=str(
                    self.endpoint(
                        queries=queries,
                    )
                )
            )
        )

        response_json = response.json()
        items = [self.model(**action) for action in response_json.get("items")]
        if len(actions) > 0:
            required = [t.cast(Action, action).slug for action in actions]
            items = [item for item in items if item.name in required]

        if len(tags) > 0:
            required_tags = [tag.app if isinstance(tag, Tag) else tag for tag in tags]
            only_important_tag = required_tags == ["important"]
            should_not_filter_using_tags = len(items) < 15 and only_important_tag
            if not should_not_filter_using_tags:
                filtered_items = [
                    item
                    for item in items
                    if any(tag in required_tags for tag in item.tags)
                ]
                if len(filtered_items) > 0 or not only_important_tag:
                    items = filtered_items

        if len(local_apps) > 0 or len(local_actions) > 0:
            local_items = self.client.local.get_action_schemas(
                apps=local_apps, actions=local_actions, tags=tags
            )
            items = [self.model(**item) for item in local_items] + items
        return items

    def execute(
        self,
        action: Action,
        params: t.Dict,
        entity_id: str = "default",
        connected_account: t.Optional[str] = None,
        session_id: t.Optional[str] = None,
        text: t.Optional[str] = None,
        auth: t.Optional[CustomAuthObject] = None,
    ) -> t.Dict:
        """
        Execute an action on the specified entity with optional connected account.

        :param action: The Action object to be executed.
        :param params: A dictionary of parameters to be passed to the action.
        :param entity_id: The unique identifier of the entity on which the action is executed.
        :param connected_account: Optional connected account ID if required for the action.
        :param session_id: ID of the current workspace session
        :return: A dictionary containing the response from the executed action.
        """
        # TOFIX: Remvoe this
        if action.is_local:
            return self.client.local.execute_action(action=action, request_data=params)

        actions = self.get(actions=[action])
        if len(actions) == 0:
            raise ComposioClientError(f"Action {action} not found")

        (action_model,) = actions
        action_req_schema = action_model.parameters.properties
        modified_params: t.Dict[str, t.Union[str, t.Dict[str, str]]] = {}
        for param, value in params.items():
            request_param_schema = action_req_schema[param]
            file_readable = request_param_schema.get("file_readable", False)
            file_uploadable = _check_file_uploadable(request_param_schema)

            if file_readable and isinstance(value, str) and os.path.isfile(value):
                with open(value, "rb") as file:
                    file_content = file.read()
                    try:
                        modified_params[param] = file_content.decode("utf-8")
                    except UnicodeDecodeError:
                        # If decoding fails, treat as binary and encode in base64
                        modified_params[param] = base64.b64encode(file_content).decode(
                            "utf-8"
                        )
            elif file_uploadable and isinstance(value, str):
                if not os.path.isfile(value):
                    raise ValueError(f"Attachment File with path `{value}` not found.")

                with open(value, "rb") as file:
                    file_content = file.read()

                modified_params[param] = {
                    "name": os.path.basename(value),
                    "content": base64.b64encode(file_content).decode("utf-8"),
                }
            else:
                modified_params[param] = value

        if action.no_auth:
            return self._raise_if_required(
                self.client.http.post(
                    url=str(self.endpoint / action.slug / "execute"),
                    json={
                        "appName": action.app,
                        "input": modified_params,
                        "text": text,
                        "sessionInfo": {
                            "sessionId": session_id,
                        },
                    },
                )
            ).json()

        if connected_account is None and auth is None:
            raise ComposioClientError(
                "`connected_account` cannot be `None` when executing "
                "an app which requires authentication"
            )

        return self._raise_if_required(
            self.client.http.post(
                url=str(self.endpoint / action.slug / "execute"),
                json={
                    "connectedAccountId": connected_account,
                    "entityId": entity_id,
                    "appName": action.app,
                    "input": modified_params,
                    "text": text,
                    "authConfig": self._serialize_auth(auth=auth),
                },
            )
        ).json()

    def request(
        self,
        connection_id: str,
        endpoint: str,
        method: str,
        body: t.Optional[t.Dict] = None,
        parameters: t.Optional[t.List[CustomAuthParameter]] = None,
    ) -> t.Dict:
        return self.client.http.post(
            url=str(self.endpoint / "proxy"),
            json={
                "connectedAccountId": connection_id,
                "body": body,
                "method": method.upper(),
                "endpoint": endpoint,
                "parameters": [
                    {
                        "in": param["in_"],
                        "name": param["name"],
                        "value": param["value"],
                    }
                    for param in parameters or []
                ],
            },
        ).json()

    @staticmethod
    def _serialize_auth(auth: t.Optional[CustomAuthObject]) -> t.Optional[t.Dict]:
        if auth is None:
            return None

        data = auth.model_dump(exclude_none=True)
        data["parameters"] = [
            {"in": d["in_"], "name": d["name"], "value": d["value"]}
            for d in data["parameters"]
        ]
        return data


class ExpectedFieldInput(BaseModel):
    name: str
    type: str

    description: str
    displayName: str
    is_secret: bool = False

    required: bool = True
    expected_from_customer: bool = True

    default: t.Optional[str] = None
    get_current_user_endpoint: t.Optional[str] = None


class IntegrationModel(BaseModel):
    """Integration data model."""

    id: str
    name: str
    authScheme: str
    createdAt: str
    updatedAt: str
    enabled: bool
    deleted: bool
    appId: str
    appName: str
    expectedInputFields: t.List[ExpectedFieldInput] = Field(default_factory=lambda: [])

    _count: t.Dict

    logo: t.Optional[str] = None
    defaultConnectorId: t.Optional[str] = None
    connections: t.Optional[t.List[t.Dict]] = None


class Integrations(Collection[IntegrationModel]):
    """
    Collection of composio integrations.
    """

    model = IntegrationModel
    endpoint = v1.integrations

    def create(
        self,
        app_id: str,
        name: t.Optional[str] = None,
        auth_mode: t.Optional["AuthSchemeType"] = None,
        auth_config: t.Optional[t.Dict[str, t.Any]] = None,
        use_composio_auth: bool = False,
        force_new_integration: bool = False,
    ) -> IntegrationModel:
        """
        Create a new integration

        :param app_id: App ID string.
        :param name: Name of the integration.
        :param auth_param: Auth mode string.
        :param auth_config: Authentication configuration.
        :param use_composio_auth: Whether to use default composio auth or not
        :return: Integration model created by the request.
        """
        request = {
            "appId": app_id,
            "useComposioAuth": use_composio_auth,
        }

        if name is not None:
            request["name"] = name

        if auth_mode is not None:
            request["authScheme"] = auth_mode

        if auth_config is not None:
            request["authConfig"] = auth_config or {}

        if force_new_integration:
            request["forceNewIntegration"] = force_new_integration

        response = self._raise_if_required(
            response=self.client.http.post(
                url=str(self.endpoint),
                json=request,
            )
        )
        return IntegrationModel(**response.json())

    def remove(self, id: str) -> None:
        self.client.http.delete(url=str(self.endpoint / id))

    @t.overload  # type: ignore
    def get(self) -> t.List[IntegrationModel]: ...

    @t.overload
    def get(self, id: t.Optional[str] = None) -> IntegrationModel: ...

    def get(
        self, id: t.Optional[str] = None
    ) -> t.Union[t.List[IntegrationModel], IntegrationModel]:
        if id is not None:
            return IntegrationModel(
                **self._raise_if_required(
                    self.client.http.get(url=str(self.endpoint / id))
                ).json()
            )
        return super().get({})

    @te.deprecated("`get_id` is deprecated, use `get(id=id)`")
    def get_by_id(
        self,
        integration_id: str,
    ) -> IntegrationModel:
        """
        Get an integration by its ID.

        :param integration_id: Integration ID string.
        :return: Integration model.
        """
        response = self._raise_if_required(
            self.client.http.get(url=str(self.endpoint / integration_id))
        )
        return IntegrationModel(**response.json())


class LogRecord(BaseModel):
    pass


class Logs(Collection[LogRecord]):
    """
    Logs endpoint.
    """

    model = LogRecord
    endpoint = v1.logs

    def push(self, record: t.Dict) -> None:
        """Push logs to composio."""
        self.client.http.post(url=str(self.endpoint), json=record)



from typing import Dict

from pydantic import Field

from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class ChwdirRequest(BaseFileRequest):
    """Request to change the current working directory."""

    path: str = Field(
        ...,
        description="The path to change the current working directory to. "
        "Can be absolute, relative to the current working directory, or use '..' to navigate up the directory tree.",
    )


class ChwdirResponse(BaseFileResponse):
    """Response to change the current working directory."""


class ChangeWorkingDirectory(LocalAction[ChwdirRequest, ChwdirResponse]):
    """
    Changes the current working directory of the file manager to the specified
    path. The commands after this action will be executed in this new directory.

    Can result in:
    - PermissionError: If the user doesn't have permission to access the directory.
    - FileNotFoundError: If the directory or any parent directory does not exist.
    - RuntimeError: If the path cannot be resolved due to a loop or other issues.
    """

    @include_cwd  # type: ignore
    def execute(self, request: ChwdirRequest, metadata: Dict) -> ChwdirResponse:
        try:
            self.filemanagers.get(request.file_manager_id).chdir(request.path)
            return ChwdirResponse()
        except PermissionError as e:
            return ChwdirResponse(error=f"Permission denied: {str(e)}")
        except FileNotFoundError as e:
            return ChwdirResponse(error=f"Directory not found: {str(e)}")
        except RuntimeError as e:
            return ChwdirResponse(error=f"Unable to resolve path: {str(e)}")
        except OSError as e:
            return ChwdirResponse(error=f"Unable to resolve path: {str(e)}")



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetTaskMembersRequest(BaseModel):
    """Request schema for `GetTaskMembers`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )


class GetTaskMembersResponse(BaseModel):
    """Response schema for `GetTaskMembers`"""

    data: t.Dict[str, t.Any]


class GetTaskMembers(OpenAPIAction):
    """View the people who have access to a task."""

    _tags = ["Members"]
    _display_name = "get_task_members"
    _request_schema = GetTaskMembersRequest
    _response_schema = GetTaskMembersResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/member"
    method = "get"
    operation_id = "Members_getTaskAccess"
    action_identifier = "/task/{task_id}/member_get"

    path_params = {"task_id": "task_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetWorkspaceEverythingLevelViewsRequest(BaseModel):
    """Request schema for `GetWorkspaceEverythingLevelViews`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )


class GetWorkspaceEverythingLevelViewsResponse(BaseModel):
    """Response schema for `GetWorkspaceEverythingLevelViews`"""

    data: t.Dict[str, t.Any]


class GetWorkspaceEverythingLevelViews(OpenAPIAction):
    """
    View the task and page views available at the Everything Level of a Workspace.
    """

    _tags = ["Views"]
    _display_name = "get_workspace_everything_level_views"
    _request_schema = GetWorkspaceEverythingLevelViewsRequest
    _response_schema = GetWorkspaceEverythingLevelViewsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/view"
    method = "get"
    operation_id = "Views_getEverythingLevel"
    action_identifier = "/team/{team_id}/view_get"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}


class views_get_everything_level(OpenAPIAction):
    """
    View the task and page views available at the Everything Level of a Workspace.<<DEPRECATED
    use get_workspace_everything_level_views>>
    """

    _tags = ["Views"]
    _display_name = "views_get_everything_level"
    _request_schema = GetWorkspaceEverythingLevelViewsRequest
    _response_schema = GetWorkspaceEverythingLevelViewsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/view"
    method = "get"
    operation_id = "Views_getEverythingLevel"
    action_identifier = "/team/{team_id}/view_get"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
Global constants for Composio SDK
"""

from pathlib import Path


ENV_COMPOSIO_API_KEY = "COMPOSIO_API_KEY"
"""
Environment variable for Composio API key
"""

ENV_COMPOSIO_LOGGING_LEVEL = "COMPOSIO_LOGGING_LEVEL"
"""
Environment variable for specifying logging level
"""

LOCAL_CACHE_DIRECTORY_NAME = ".composio"
"""
Local cache directory name for composio CLI
"""

LOCAL_CACHE_DIRECTORY = Path.home() / LOCAL_CACHE_DIRECTORY_NAME
"""
Path to local caching directory.
"""

LOCAL_OUTPUT_FILE_DIRECTORY_NAME = "output"
"""
Local output file directory name for composio tools
"""

USER_DATA_FILE_NAME = "user_data.json"
"""
Filename for storing user data.
"""

DEFAULT_ENTITY_ID = "default"
"""
Default entity ID value.
"""

DEFAULT_BASE_URL = "https://backend.composio.dev/api"
"""
Base URL for composio API server.
"""

ENV_COMPOSIO_BASE_URL = "COMPOSIO_BASE_URL"
"""
Environment variable for Composio API server base URL
"""

BASE_URL_PROD = "https://backend.composio.dev/api"
"""
Base URL for production server.
"""

WEB_URL_PROD = "https://app.composio.dev"
"""
Web URL for production server.
"""

BASE_URL_STAGING = "https://staging-backend.composio.dev/api"
"""
Base URL for staging server.
"""

WEB_URL_STAGING = "https://hermes-frontend-git-master-composio.vercel.app"
"""
Web URL for staging server.
"""

BASE_URL_LOCAL = "http://localhost:9900/api"
"""
Base URL for local server.
"""

WEB_URL_LOCAL = "http://localhost:3000"
"""
Web URL for local server.
"""

BASE_URL_TO_PROD_MAPPING = {
    BASE_URL_PROD: WEB_URL_PROD,
    BASE_URL_STAGING: WEB_URL_STAGING,
    BASE_URL_LOCAL: WEB_URL_LOCAL,
}
"""
Composio API server base url -> web url mappings.
"""

PUSHER_KEY = "ff9f18c208855d77a152"
"""
API Key for Pusher subscriptions.
"""

PUSHER_CLUSTER = "mt1"
"""
Name of the pusher cluster.
"""



import hashlib
import os
import sys
import time
from functools import wraps
from typing import Optional


def is_test_file(file_path: str) -> bool:
    """
    Determines whether a file is a test-related file or not.

    Args:
        file_path (str): The path of the file to be checked.

    Returns:
        bool: True if the file is a test-related file, False otherwise.

    Criteria to decide whether a file is a test-related file or not:
    - If the file name contains "_test" or "test_".
    - If the file path contains the directory "tests/".

    Reference: https://docs.pytest.org/en/7.1.x/explanation/goodpractices.html#test-discovery
    """
    file_path_name = os.path.basename(file_path)
    if "_test" in file_path_name:
        return True
    if "test_" in file_path_name:
        return True
    if "tests/" in file_path:
        return True
    return False


def find_python_files(
    directory: str,
    filter_test_files: bool = True,
    filter_out_unreadable_files: bool = True,
) -> list:
    """
    Fetch all python files in the given directory.

    Args:
        directory (str): The directory to search for python files.
        filter_test_files (bool, optional): Whether to filter out test files. Defaults to True.
        filter_out_unreadable_files (bool, optional): Whether to filter out unreadable files. Defaults to True.

    Returns:
        list: A list of absolute paths to the python files found.
    """
    directory = os.path.normpath(os.path.abspath(directory))
    python_files = [
        os.path.normpath(os.path.abspath(os.path.join(root, file)))
        for root, _, files in os.walk(directory)
        for file in files
        if file.endswith(".py")
    ]
    print(f"[ToolObj] Total number of python files found: {len(python_files)}")

    if filter_test_files:
        python_files = [
            file
            for file in python_files
            if not is_test_file(file.replace(directory, ""))
        ]
        print(
            f"[ToolObj] Total number of python files after filtering out test files: {len(python_files)}"
        )

    if filter_out_unreadable_files:
        readable_files = []
        for file in python_files:
            try:
                with open(file, "r", encoding="utf-8") as f:
                    f.read()
                readable_files.append(file)
            except Exception:
                pass
        python_files = readable_files
        print(
            f"Total number of python files found (after removing test-files + unreadable files): {len(python_files)}"
        )

    return python_files


def get_virtual_env_name() -> Optional[str]:
    """
    Infers the name of the virtual environment from the path to the Python executable.

    Returns:
        Optional[str]: The name of the virtual environment, or None if not in a virtual environment.
    """
    executable_path = sys.executable
    path_parts = executable_path.split(os.sep)

    try:
        # Look for a directory name typically associated with a virtual environment
        env_index = path_parts.index("bin") - 1
        return path_parts[env_index]
    except ValueError:
        # If "bin" is not found in the path, we're likely not in a virtual environment
        return None


def fetch_hash(message) -> str:
    """
    Calculates the SHA256 hash of the given message.

    Args:
        message (str): The message to calculate the hash for.

    Returns:
        str: The hexadecimal representation of the SHA256 hash.
    """
    encoded_message = message.encode()

    hash_object = hashlib.sha256()
    hash_object.update(encoded_message)

    hex_dig = hash_object.hexdigest()
    return hex_dig


def retry_handler(max_attempts=3, delay=1):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_attempts:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempts += 1
                    if attempts == max_attempts:
                        raise RuntimeError(
                            f"Failed after {max_attempts} attempts: {e}"
                        ) from e
                    time.sleep(delay)
            raise RuntimeError(f"Failed after {max_attempts} attempts")

        return wrapper

    return decorator



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class EditSpaceTagRequest(BaseModel):
    """Request schema for `EditSpaceTag`"""

    space_id: int = Field(
        ...,
        alias="space_id",
        description="",
    )
    tag_name: str = Field(
        ...,
        alias="tag_name",
        description="",
    )
    tag_fg_color: str = Field(
        default=...,
        alias="tag__fg_color",
        description="Fg Color",
    )
    tag_bg_color: str = Field(
        default=...,
        alias="tag__bg_color",
        description="Bg Color",
    )


class EditSpaceTagResponse(BaseModel):
    """Response schema for `EditSpaceTag`"""

    data: t.Dict[str, t.Any]


class EditSpaceTag(OpenAPIAction):
    """Update a task Tag."""

    _tags = ["Tags"]
    _display_name = "edit_space_tag"
    _request_schema = EditSpaceTagRequest
    _response_schema = EditSpaceTagResponse

    url = "https://api.clickup.com/api/v2"
    path = "/space/{space_id}/tag/{tag_name}"
    method = "put"
    operation_id = "Tags_updateSpaceTag"
    action_identifier = "/space/{space_id}/tag/{tag_name}_put"

    path_params = {"space_id": "space_id", "tag_name": "tag_name"}
    query_params = {}
    header_params = {}
    request_params = {
        "tag": {
            "__alias": "tag",
            "fg_color": {"__alias": "fg_color"},
            "bg_color": {"__alias": "bg_color"},
        }
    }

    aliases = {"tag": "76f59ca3"}



"""
Action for getting details of the current webpage.
"""

from typing import Any, Dict, Optional

from pydantic import Field

from composio.tools.env.browsermanager.manager import BrowserManager
from composio.tools.local.browsertool.actions.base_action import (
    BaseBrowserAction,
    BaseBrowserRequest,
    BaseBrowserResponse,
)


class GetPageDetailsRequest(BaseBrowserRequest):
    """Request schema for getting page details."""

    include_accessibility: bool = Field(
        default=True, description="Whether to include accessibility snapshot"
    )


class GetPageDetailsResponse(BaseBrowserResponse):
    """Response schema for getting page details."""

    details: Optional[Dict[str, Any]] = Field(
        default=None, description="Details of the current page"
    )
    success: bool = Field(
        default=False, description="Whether the action was successful"
    )


class GetPageDetails(BaseBrowserAction[GetPageDetailsRequest, GetPageDetailsResponse]):
    """
    Get details of the current webpage.

    This action retrieves various details about the currently loaded webpage in the browser.
    It can include information such as the page title, URL, meta tags, and optionally,
    an accessibility snapshot.
    """

    display_name = "GetPageDetails"
    _request_schema = GetPageDetailsRequest
    _response_schema = GetPageDetailsResponse

    def execute_on_browser_manager(
        self, browser_manager: BrowserManager, request: GetPageDetailsRequest  # type: ignore
    ) -> GetPageDetailsResponse:
        """Execute the get page details action."""
        page_details = browser_manager.get_page_details()
        if not request.include_accessibility:
            page_details.pop("page_details", None)
        return GetPageDetailsResponse(success=True, details=page_details)



from .add_dependency import AddDependency
from .add_guest_to_folder import AddGuestToFolder
from .add_guest_to_list import AddGuestToList
from .add_guest_to_task import AddGuestToTask
from .add_tag_to_task import AddTagToTask
from .add_tags_from_time_entries import AddTagsFromTimeEntries
from .add_task_link import AddTaskLink
from .add_task_to_list import AddTaskToList
from .change_tag_names_from_time_entries import ChangeTagNamesFromTimeEntries
from .create_a_time_entry import CreateATimeEntry
from .create_chat_view_comment import CreateChatViewComment
from .create_checklist import CreateChecklist, task_checklists_create_new_checklist
from .create_checklist_item import CreateChecklistItem
from .create_folder import CreateFolder, folders_create_new_folder
from .create_folder_view import CreateFolderView
from .create_folderless_list import CreateFolderlessList
from .create_goal import CreateGoal
from .create_key_result import CreateKeyResult
from .create_list import CreateList
from .create_list_comment import CreateListComment
from .create_list_view import CreateListView
from .create_space import CreateSpace
from .create_space_tag import CreateSpaceTag
from .create_space_view import CreateSpaceView
from .create_task import CreateTask
from .create_task_attachment import (
    CreateTaskAttachment,
    attachments_upload_file_to_task_as_attachment,
)
from .create_task_comment import CreateTaskComment
from .create_task_from_template import CreateTaskFromTemplate
from .create_team import CreateTeam, teams_user_groups_create_team
from .create_webhook import CreateWebhook
from .create_workspace_everything_level_view import CreateWorkspaceEverythingLevelView
from .delete_a_time_entry import DeleteATimeEntry
from .delete_checklist import DeleteChecklist
from .delete_checklist_item import DeleteChecklistItem
from .delete_comment import DeleteComment
from .delete_dependency import DeleteDependency
from .delete_folder import DeleteFolder
from .delete_goal import DeleteGoal
from .delete_key_result import DeleteKeyResult
from .delete_list import DeleteList
from .delete_space import DeleteSpace
from .delete_space_tag import DeleteSpaceTag
from .delete_task import DeleteTask
from .delete_task_link import DeleteTaskLink
from .delete_team import DeleteTeam
from .delete_time_tracked import DeleteTimeTracked
from .delete_view import DeleteView
from .delete_webhook import DeleteWebhook
from .edit_checklist import EditChecklist
from .edit_checklist_item import EditChecklistItem
from .edit_guest_on_workspace import EditGuestOnWorkspace
from .edit_key_result import EditKeyResult
from .edit_space_tag import EditSpaceTag
from .edit_time_tracked import EditTimeTracked
from .edit_user_on_workspace import EditUserOnWorkspace
from .get_access_token import GetAccessToken, authorization_get_access_token
from .get_accessible_custom_fields import GetAccessibleCustomFields
from .get_all_tags_from_time_entries import GetAllTagsFromTimeEntries
from .get_authorized_teams_workspaces import (
    GetAuthorizedTeamsWorkspaces,
    authorization_get_work_space_list,
)
from .get_authorized_user import GetAuthorizedUser, authorization_view_account_details
from .get_bulk_tasks_time_in_status import GetBulkTasksTimeInStatus
from .get_chat_view_comments import GetChatViewComments
from .get_custom_roles import GetCustomRoles
from .get_custom_task_types import GetCustomTaskTypes
from .get_filtered_team_tasks import GetFilteredTeamTasks
from .get_folder import GetFolder, folders_get_folder_content
from .get_folder_views import GetFolderViews
from .get_folderless_lists import GetFolderlessLists
from .get_folders import GetFolders, folders_get_contents_of
from .get_goal import GetGoal
from .get_goals import GetGoals
from .get_guest import GetGuest
from .get_list import GetList
from .get_list_comments import GetListComments
from .get_list_members import GetListMembers, members_get_list_users
from .get_list_views import GetListViews
from .get_lists import GetLists, lists_get_folder_lists
from .get_running_time_entry import GetRunningTimeEntry
from .get_singular_time_entry import GetSingularTimeEntry
from .get_space import GetSpace, spaces_get_details
from .get_space_tags import GetSpaceTags
from .get_space_views import GetSpaceViews, views_space_views_get
from .get_spaces import GetSpaces, spaces_get_space_details
from .get_task import GetTask, tasks_get_task_details
from .get_task_comments import GetTaskComments
from .get_task_members import GetTaskMembers
from .get_task_s_time_in_status import GetTaskSTimeInStatus
from .get_task_templates import GetTaskTemplates
from .get_tasks import GetTasks, tasks_get_list_tasks
from .get_teams import GetTeams
from .get_time_entries_within_a_date_range import GetTimeEntriesWithinADateRange
from .get_time_entry_history import GetTimeEntryHistory
from .get_tracked_time import GetTrackedTime
from .get_user import GetUser
from .get_view import GetView
from .get_view_tasks import GetViewTasks
from .get_webhooks import GetWebhooks
from .get_workspace_everything_level_views import (
    GetWorkspaceEverythingLevelViews,
    views_get_everything_level,
)
from .get_workspace_plan import GetWorkspacePlan, teams_work_spaces_get_work_space_plan
from .get_workspace_seats import (
    GetWorkspaceSeats,
    teams_work_spaces_get_work_space_seats,
)
from .invite_guest_to_workspace import InviteGuestToWorkspace
from .invite_user_to_workspace import InviteUserToWorkspace
from .remove_custom_field_value import RemoveCustomFieldValue
from .remove_guest_from_folder import RemoveGuestFromFolder
from .remove_guest_from_list import RemoveGuestFromList
from .remove_guest_from_task import RemoveGuestFromTask
from .remove_guest_from_workspace import RemoveGuestFromWorkspace
from .remove_tag_from_task import RemoveTagFromTask
from .remove_tags_from_time_entries import RemoveTagsFromTimeEntries
from .remove_task_from_list import RemoveTaskFromList
from .remove_user_from_workspace import RemoveUserFromWorkspace
from .set_custom_field_value import SetCustomFieldValue
from .shared_hierarchy import SharedHierarchy
from .start_a_time_entry import StartATimeEntry
from .stop_a_time_entry import StopATimeEntry
from .track_time import TrackTime
from .update_a_time_entry import UpdateATimeEntry
from .update_comment import UpdateComment
from .update_folder import UpdateFolder
from .update_goal import UpdateGoal
from .update_list import UpdateList
from .update_space import UpdateSpace
from .update_task import UpdateTask
from .update_team import UpdateTeam
from .update_view import UpdateView
from .update_webhook import UpdateWebhook


__all__ = (
    "CreateTaskAttachment",
    "attachments_upload_file_to_task_as_attachment",
    "GetAccessToken",
    "authorization_get_access_token",
    "GetAuthorizedUser",
    "authorization_view_account_details",
    "GetAuthorizedTeamsWorkspaces",
    "authorization_get_work_space_list",
    "CreateChecklist",
    "task_checklists_create_new_checklist",
    "EditChecklist",
    "DeleteChecklist",
    "CreateChecklistItem",
    "EditChecklistItem",
    "DeleteChecklistItem",
    "GetTaskComments",
    "CreateTaskComment",
    "GetChatViewComments",
    "CreateChatViewComment",
    "GetListComments",
    "CreateListComment",
    "UpdateComment",
    "DeleteComment",
    "GetAccessibleCustomFields",
    "SetCustomFieldValue",
    "RemoveCustomFieldValue",
    "AddDependency",
    "DeleteDependency",
    "AddTaskLink",
    "DeleteTaskLink",
    "GetFolders",
    "folders_get_contents_of",
    "CreateFolder",
    "folders_create_new_folder",
    "GetFolder",
    "folders_get_folder_content",
    "UpdateFolder",
    "DeleteFolder",
    "GetGoals",
    "CreateGoal",
    "GetGoal",
    "UpdateGoal",
    "DeleteGoal",
    "CreateKeyResult",
    "EditKeyResult",
    "DeleteKeyResult",
    "InviteGuestToWorkspace",
    "GetGuest",
    "EditGuestOnWorkspace",
    "RemoveGuestFromWorkspace",
    "AddGuestToTask",
    "RemoveGuestFromTask",
    "AddGuestToList",
    "RemoveGuestFromList",
    "AddGuestToFolder",
    "RemoveGuestFromFolder",
    "GetLists",
    "lists_get_folder_lists",
    "CreateList",
    "GetFolderlessLists",
    "CreateFolderlessList",
    "GetList",
    "UpdateList",
    "DeleteList",
    "AddTaskToList",
    "RemoveTaskFromList",
    "GetTaskMembers",
    "GetListMembers",
    "members_get_list_users",
    "GetCustomRoles",
    "SharedHierarchy",
    "GetSpaces",
    "spaces_get_space_details",
    "CreateSpace",
    "GetSpace",
    "spaces_get_details",
    "UpdateSpace",
    "DeleteSpace",
    "GetSpaceTags",
    "CreateSpaceTag",
    "EditSpaceTag",
    "DeleteSpaceTag",
    "AddTagToTask",
    "RemoveTagFromTask",
    "GetTasks",
    "tasks_get_list_tasks",
    "CreateTask",
    "GetTask",
    "tasks_get_task_details",
    "UpdateTask",
    "DeleteTask",
    "GetFilteredTeamTasks",
    "GetTaskSTimeInStatus",
    "GetBulkTasksTimeInStatus",
    "GetTaskTemplates",
    "CreateTaskFromTemplate",
    "GetWorkspaceSeats",
    "teams_work_spaces_get_work_space_seats",
    "GetWorkspacePlan",
    "teams_work_spaces_get_work_space_plan",
    "CreateTeam",
    "teams_user_groups_create_team",
    "GetCustomTaskTypes",
    "UpdateTeam",
    "DeleteTeam",
    "GetTeams",
    "GetTrackedTime",
    "TrackTime",
    "EditTimeTracked",
    "DeleteTimeTracked",
    "GetTimeEntriesWithinADateRange",
    "CreateATimeEntry",
    "GetSingularTimeEntry",
    "DeleteATimeEntry",
    "UpdateATimeEntry",
    "GetTimeEntryHistory",
    "GetRunningTimeEntry",
    "RemoveTagsFromTimeEntries",
    "GetAllTagsFromTimeEntries",
    "AddTagsFromTimeEntries",
    "ChangeTagNamesFromTimeEntries",
    "StartATimeEntry",
    "StopATimeEntry",
    "InviteUserToWorkspace",
    "GetUser",
    "EditUserOnWorkspace",
    "RemoveUserFromWorkspace",
    "GetWorkspaceEverythingLevelViews",
    "views_get_everything_level",
    "CreateWorkspaceEverythingLevelView",
    "GetSpaceViews",
    "views_space_views_get",
    "CreateSpaceView",
    "GetFolderViews",
    "CreateFolderView",
    "GetListViews",
    "CreateListView",
    "GetView",
    "UpdateView",
    "DeleteView",
    "GetViewTasks",
    "GetWebhooks",
    "CreateWebhook",
    "UpdateWebhook",
    "DeleteWebhook",
)



import json
import os
from abc import abstractmethod
from typing import Dict, List, Optional

from composio.tools.local.codeanalysis.constants import CODE_MAP_CACHE, FQDN_FILE


class BaseCodeAnalysisAction:
    def __init__(self):
        self.fqdn_cache_file = None
        self.all_fqdns_df = None
        self.fqdn_index = None

    @abstractmethod
    def execute(self, request, metadata):
        pass

    def load_fqdn_cache(self, repo_name: str):
        self.fqdn_cache_file = os.path.join(CODE_MAP_CACHE, repo_name, FQDN_FILE)
        if not os.path.exists(self.fqdn_cache_file):
            raise FileNotFoundError(
                f"FQDN cache file not found: {self.fqdn_cache_file}"
            )

        with open(self.fqdn_cache_file, "r", encoding="utf-8") as f:
            self.all_fqdns_df = json.load(f)

        self.fqdn_index = {
            fqdn_obj["global_fqdn"]: fqdn_obj
            for possible_fqdns in self.all_fqdns_df.values()
            for fqdn_obj in possible_fqdns
        }

    def get_matching_items(
        self,
        query_name: Optional[str],
        item_type: str,
        parent_fqdns: Optional[List[str]] = None,
    ) -> List[str]:
        if not self.fqdn_index:
            raise ValueError("FQDN index not loaded")

        def matches_query(curr_fqdn: str) -> bool:
            return (
                query_name is None
                or query_name == curr_fqdn.split(".")[-1]
                or query_name == curr_fqdn
            )

        matching_fqdns = [
            curr_fqdn
            for curr_fqdn, curr_fqdn_elem in self.fqdn_index.items()
            if curr_fqdn_elem["global_type"] == item_type
            and matches_query(curr_fqdn)
            and (parent_fqdns is None or curr_fqdn_elem["parent_fqdn"] in parent_fqdns)
        ]

        return matching_fqdns

    def fetch_relevant_details(self, relevant_fqdn: str, repo_path: str) -> List[Dict]:
        from composio.tools.local.codeanalysis import (  # pylint: disable=import-outside-toplevel
            lsp_helper,
        )

        if self.fqdn_index is None:
            raise ValueError("FQDN index not loaded")
        elem_fqdn = self.fqdn_index[relevant_fqdn]
        elem = lsp_helper.fetch_relevant_elem(
            elem_fqdn["global_module"],
            repo_path,
            elem_fqdn["global_fqdn"],
            elem_fqdn["global_type"],
        )
        if isinstance(elem, list):
            return [x.__dict__ for x in elem]
        raise ValueError("Expected a list of elements")

    def get_item_results(self, matching_fqdns: List[str], repo_path: str) -> List[Dict]:
        matching_fqdn_elems_df = {
            k: self.fetch_relevant_details(k, repo_path) for k in matching_fqdns
        }
        results = []
        for _val in matching_fqdn_elems_df.values():
            if isinstance(_val, list):
                results.extend(_val)
            else:
                results.append(_val)
        return results


class MethodAnalysisAction(BaseCodeAnalysisAction):
    @abstractmethod
    def execute(self, request, metadata):
        pass

    def get_method_artefacts(
        self, query_class_name: Optional[str], query_method_name: str, repo_path: str
    ) -> Dict:
        matching_fqdns_class = self.get_matching_items(query_class_name, "class")
        if query_class_name is not None:
            matching_fqdns_func = self.get_matching_items(
                query_method_name, "function", matching_fqdns_class
            )
        else:
            matching_fqdns_func = self.get_matching_items(query_method_name, "function")
        func_results = self.get_item_results(matching_fqdns_func, repo_path)
        if query_class_name is not None:
            filtered_func_results = self.filter_function_results(
                func_results, query_class_name, matching_fqdns_class, repo_path
            )
        else:
            filtered_func_results = func_results

        return self.format_method_results(filtered_func_results)

    def filter_function_results(
        self,
        func_results: List[Dict],
        query_class_name: Optional[str],
        matching_fqdns_class: List[str],
        repo_path: str,
    ) -> List[Dict]:
        filtered_results = []
        for func_res in func_results:
            parent_class = func_res["parent_class"]
            if parent_class is None:
                if query_class_name is None:
                    filtered_results.append(func_res)
            else:
                if self.is_function_in_class(func_res, matching_fqdns_class, repo_path):
                    filtered_results.append(func_res)
        return filtered_results

    def is_function_in_class(
        self, func_res: Dict, matching_fqdns_class: List[str], repo_path: str
    ) -> bool:
        matching_class_elems = {
            k: self.fetch_relevant_details(k, repo_path) for k in matching_fqdns_class
        }
        for class_elem in matching_class_elems.values():
            all_members = self.get_all_members(class_elem)
            if func_res["full_name"] in all_members:
                return True
        return False

    @staticmethod
    def get_all_members(class_elem: List[Dict]) -> List[str]:
        all_members = []
        for elem in class_elem:
            all_members.extend(elem["member_functions"])
        return all_members

    @staticmethod
    def format_method_results(method_results: List[Dict]) -> Dict:
        if not method_results:
            return {
                "signature_ans": "No matching methods found.",
                "body_ans": "No matching methods found.",
            }

        signature_ans = []
        body_ans = []

        for idx, func in enumerate(method_results):
            result_header = f"## Details about shortlisted result ID {idx}:\n"
            function_details = func["res_fetch_function_stuff"]
            function_body = "\n```python\n"
            for line_num, line in enumerate(
                func["definition_body"].splitlines(), start=func["start_line"] + 1
            ):
                function_body += f"{line_num}: {line}\n"
            function_body += "```"

            signature_ans.append(result_header + function_details)
            body_ans.append(result_header + function_details + function_body)

        return {
            "signature_ans": "\n".join(signature_ans),
            "body_ans": "\n".join(body_ans),
        }



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetBulkTasksTimeInStatusRequest(BaseModel):
    """Request schema for `GetBulkTasksTimeInStatus`"""

    task_ids: str = Field(
        ...,
        alias="task_ids",
        description=(
            "Include this parameter once per `task_id`. You can include up to 100 task "
            "ids per request. For example: `task_ids=3cuh&task_ids=g4fs` "
        ),
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )


class GetBulkTasksTimeInStatusResponse(BaseModel):
    """Response schema for `GetBulkTasksTimeInStatus`"""

    data: t.Dict[str, t.Any]


class GetBulkTasksTimeInStatus(OpenAPIAction):
    """View how long two or more tasks have been in each status."""

    _tags = ["Tasks"]
    _display_name = "get_bulk_tasks_time_in_status"
    _request_schema = GetBulkTasksTimeInStatusRequest
    _response_schema = GetBulkTasksTimeInStatusResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/bulk_time_in_status/task_ids"
    method = "get"
    operation_id = "Tasks_getTimeInStatusBulk"
    action_identifier = "/task/bulk_time_in_status/task_ids_get"

    path_params = {}
    query_params = {
        "task_ids": "task_ids",
        "custom_task_ids": "custom_task_ids",
        "team_id": "team_id",
    }
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class RemoveTagsFromTimeEntriesRequest(BaseModel):
    """Request schema for `RemoveTagsFromTimeEntries`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    tags: t.List[dict] = Field(
        default=...,
        alias="tags",
        description="",
    )
    time_entry_ids: t.List[str] = Field(
        default=...,
        alias="time_entry_ids",
        description="",
    )


class RemoveTagsFromTimeEntriesResponse(BaseModel):
    """Response schema for `RemoveTagsFromTimeEntries`"""

    data: t.Dict[str, t.Any]


class RemoveTagsFromTimeEntries(OpenAPIAction):
    """
    Remove labels from time entries. This does not remove the label from a Workspace.
    """

    _tags = ["Time Tracking"]
    _display_name = "remove_tags_from_time_entries"
    _request_schema = RemoveTagsFromTimeEntriesRequest
    _response_schema = RemoveTagsFromTimeEntriesResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/time_entries/tags"
    method = "delete"
    operation_id = "TimeTracking_removeTagsFromTimeEntries"
    action_identifier = "/team/{team_id}/time_entries/tags_delete"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "tags": {"__alias": "tags"},
        "time_entry_ids": {"__alias": "time_entry_ids"},
    }

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class UpdateCommentRequest(BaseModel):
    """Request schema for `UpdateComment`"""

    comment_id: int = Field(
        ...,
        alias="comment_id",
        description="",
    )
    comment_text: str = Field(
        default=...,
        alias="comment_text",
        description="Comment Text",
    )
    assignee: int = Field(
        default=...,
        alias="assignee",
        description="Assignee",
    )
    resolved: bool = Field(
        default=...,
        alias="resolved",
        description="",
    )


class UpdateCommentResponse(BaseModel):
    """Response schema for `UpdateComment`"""

    data: t.Dict[str, t.Any]


class UpdateComment(OpenAPIAction):
    """
    Replace the content of a task comment, assign a comment, and mark a comment
    as resolved.
    """

    _tags = ["Comments"]
    _display_name = "update_comment"
    _request_schema = UpdateCommentRequest
    _response_schema = UpdateCommentResponse

    url = "https://api.clickup.com/api/v2"
    path = "/comment/{comment_id}"
    method = "put"
    operation_id = "Comments_updateTaskComment"
    action_identifier = "/comment/{comment_id}_put"

    path_params = {"comment_id": "comment_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "comment_text": {"__alias": "comment_text"},
        "assignee": {"__alias": "assignee"},
        "resolved": {"__alias": "resolved"},
    }

    aliases = {}



import re
from typing import Dict, List, Tuple

from tree_sitter import Node, Parser, Tree
from tree_sitter_languages import get_parser


class SpanRelated:
    """Class containing static methods for span-related operations"""

    @staticmethod
    def convert_to_tuple_format(
        original_span: List[List[int]],
    ) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """
        Converts the span to a tuple format.

        Args:
            original_span (List[List[int]]): The original span in list format.

        Returns:
            Tuple[Tuple[int, int], Tuple[int, int]]: The span in tuple format.

        Raises:
            ValueError: If the span format is invalid.
        """
        if len(original_span) != 2 or any(len(point) != 2 for point in original_span):
            raise ValueError("Invalid span format")
        return (
            (original_span[0][0], original_span[0][1]),
            (original_span[1][0], original_span[1][1]),
        )

    @staticmethod
    def has_span_overlap(
        span1: Tuple[Tuple[int, int], Tuple[int, int]],
        span2: Tuple[Tuple[int, int], Tuple[int, int]],
    ) -> bool:
        """
        Returns True if the spans overlap, else False.

        Args:
            span1 (Tuple[Tuple[int, int], Tuple[int, int]]): The first span.
            span2 (Tuple[Tuple[int, int], Tuple[int, int]]): The second span.

        Returns:
            bool: True if the spans overlap, else False.
        """
        span1, span2 = map(SpanRelated.convert_to_tuple_format, (span1, span2))  # type: ignore
        return not (span1[1] < span2[0] or span1[0] > span2[1])

    @staticmethod
    def does_span_contain(parent_span: Tuple, child_span: Tuple) -> bool:
        """
        Returns True if parent_span contains child_span, else False.

        Args:
            parent_span (Tuple[Tuple[int, int], Tuple[int, int]]): The parent span.
            child_span (Tuple[Tuple[int, int], Tuple[int, int]]): The child span.

        Returns:
            bool: True if parent_span contains child_span, else False.
        """
        parent_span, child_span = map(
            SpanRelated.convert_to_tuple_format, (parent_span, child_span)  # type: ignore
        )
        return (parent_span[0] <= child_span[0]) and (parent_span[1] >= child_span[1])


def fetch_tree(parser: Parser, code_use: str) -> Tree:
    """Parses the given code and returns the syntax tree"""
    tree = parser.parse(bytes(code_use, "utf8"))
    return tree


def fetch_relevant_body(node: Node, test_code_str: str) -> str:
    """
    Fetches the relevant body of code for a given node.

    Args:
        node (Node): The node for which the relevant body of code is to be fetched.
        test_code_str (str): The string containing the entire code.

    Returns:
        str: The relevant body of code for the given node.
    """
    start_line, start_col = node.start_point
    end_line, end_col = node.end_point
    all_lines = test_code_str.split("\n")

    if start_line == end_line:
        return all_lines[start_line][start_col:end_col]

    relevant_body = [all_lines[start_line][start_col:]]
    relevant_body.extend(all_lines[start_line + 1 : end_line])  # noqa: E203
    relevant_body.append(all_lines[end_line][:end_col])
    return "\n".join(relevant_body)


def fetch_type_nodes(node: Node, desired_types_list: List[str]) -> List[Node]:
    """Fetches nodes of the desired types from the syntax tree"""
    relevant_nodes: List[Node] = []

    if node.type in desired_types_list:
        relevant_nodes.append(node)

    for (
        curr_child
    ) in (
        node.children
    ):  # Recursively fetch nodes of the desired types from the children
        relevant_nodes += fetch_type_nodes(curr_child, desired_types_list)

    return relevant_nodes  # Return the list of relevant nodes


def fetch_nodes_of_type(file_path: str, types_allowed: List[str]) -> List[Dict]:
    """
    Fetches the locations of type annotations in the file in the form [{"node_text", "span"}].

    Args:
        file_path (str): The path to the file to be analyzed.
        types_allowed (List[str]): A list of node types to be fetched.

    Returns:
        List[Dict]: A list of dictionaries containing node details.
    """
    parser = get_parser("python")
    with open(file_path, "r", encoding="utf-8") as file:
        test_code_str = file.read()

    root_node = fetch_tree(parser, test_code_str).root_node
    generic_type_nodes = fetch_type_nodes(root_node, types_allowed)
    node_list = format_node_list(nodes=generic_type_nodes, file_body=test_code_str)
    return node_list


def format_node_list(nodes: List[Node], file_body: str) -> List[Dict]:
    """
    Formats the list of nodes with their details.

    Args:
        nodes (List[Node]): A list of nodes to be formatted.
        file_body (str): The content of the file as a string.

    Returns:
        list: A list of dictionaries containing formatted node details.
    """
    formatted_identifier_nodes = []
    for node in nodes:
        formatted_identifier_nodes.append(
            {
                "node_obj": node,
                "node_txt": fetch_relevant_body(node, file_body),
                "start_point": node.start_point,
                "end_point": node.end_point,
            }
        )

    for node in formatted_identifier_nodes:
        node["start_point"] = (
            node["start_point"][0] + 1,
            node["start_point"][1],
        )
        node["end_point"] = (
            node["end_point"][0] + 1,
            node["end_point"][1],
        )
        node["span"] = SpanRelated.convert_to_tuple_format(
            (node["start_point"], node["end_point"])  # type: ignore
        )
        del node["start_point"]
        del node["end_point"]

    formatted_identifier_nodes.sort(key=lambda x: x["span"])
    return formatted_identifier_nodes


def fetch_class_and_function_nodes_defn_identifiers(file_path: str) -> List[Dict]:
    """
    Fetches the class and function definition identifiers from the file.

    Args:
        file_path (str): The path to the file.

    Returns:
        List[Dict]: A list of dictionaries containing formatted node details.
    """
    # Fetch function and class definition nodes
    wanted_definition_nodes = fetch_nodes_of_type(
        file_path, ["function_definition", "class_definition"]
    )

    # Read the code from the file
    with open(file_path, "r", encoding="utf-8") as file:
        file_body = file.read()

    # Find the identifier nodes in the children of the definition nodes
    identifier_nodes = []
    for node in wanted_definition_nodes:
        curr_identifier_nodes = [
            x for x in node["node_obj"].children if x.type == "identifier"
        ]
        if len(curr_identifier_nodes) != 1:
            raise ValueError(
                f"Expected exactly one identifier node, found: {len(curr_identifier_nodes)}"
            )
        identifier_nodes.append(curr_identifier_nodes[0])

    # Format the list of identifier nodes
    formatted_identifier_nodes = format_node_list(identifier_nodes, file_body)

    return formatted_identifier_nodes


def find_left_side_identifiers_of_assignments(file_path: str) -> List[Dict]:
    """
    Fetches the left side identifiers of assignments from the file.

    Args:
        file_path (str): The path to the file.

    Returns:
        List[Dict]: A list of dictionaries containing formatted node details.
    """
    # Fetch assignment nodes
    wanted_definition_nodes = fetch_nodes_of_type(file_path, ["assignment"])

    # Read the code from the file
    with open(file_path, "r", encoding="utf-8") as file:
        file_body = file.read()

    identifier_nodes = []
    # Find the identifier nodes in the children of the assignment nodes
    for node in wanted_definition_nodes:
        node_obj = node["node_obj"]
        if len(node_obj.children) == 0:
            continue
        if node_obj.children[0].type == "identifier":
            identifier_nodes.append(node_obj.children[0])

    # Format the list of identifier nodes
    formatted_identifier_nodes = format_node_list(identifier_nodes, file_body)

    return formatted_identifier_nodes


def fetch_entity_artifacts(entity_body: str, entity_type: str) -> Dict[str, str]:
    """
    Fetches artifacts (signature, block, docstring) for a given entity (class or function).

    Args:
        entity_body (str): The body of the entity to analyze.
        entity_type (str): The type of the entity ("class" or "function").

    Returns:
        Dict[str, str]: A dictionary containing the entity's signature, block, and docstring.

    Raises:
        ValueError: If the entity_type is invalid or if the expected nodes are not found.
    """
    if entity_type not in ["class", "function"]:
        raise ValueError("Invalid entity_type. Must be 'class' or 'function'.")

    parser = get_parser("python")
    tree = fetch_tree(parser, entity_body)
    entity_node = tree.root_node

    entity_defn_nodes = fetch_type_nodes(entity_node, [f"{entity_type}_definition"])
    if not entity_defn_nodes:
        raise ValueError(f"No {entity_type} definition found in the given body.")

    entity_root_node = min(entity_defn_nodes, key=lambda x: x.start_point)
    block_nodes = fetch_type_nodes(entity_root_node, ["block"])

    if not block_nodes:
        raise ValueError(f"No block found in the {entity_type} definition.")

    entity_block_node = min(block_nodes, key=lambda x: x.start_point)

    entity_root_node_txt = fetch_relevant_body(entity_root_node, entity_body)
    entity_block_node_txt = fetch_relevant_body(entity_block_node, entity_body)

    signature = entity_root_node_txt.split(entity_block_node_txt)[0]
    docstring = ""

    if entity_block_node.children:
        first_child = entity_block_node.children[0]
        if first_child.type == "expression_statement" and first_child.children:
            first_child_child = first_child.children[0]
            if first_child_child.type == "string":
                docstring = fetch_relevant_body(first_child_child, entity_body)

    return {
        "signature": re.sub(r"\s+", " ", signature).strip(),
        "block": entity_block_node_txt.strip(),
        "docstring": docstring.strip(),
    }



"""
Add new integration.

Usage:
    composio add [options]
"""

import typing as t
import webbrowser

import click

from composio.cli.context import Context, ensure_login, pass_context
from composio.cli.utils.decorators import handle_exceptions, pass_entity_id
from composio.cli.utils.helpfulcmd import HelpfulCmd
from composio.client import Composio, Entity
from composio.client.collections import (
    AppAuthScheme,
    AppModel,
    AuthSchemeField,
    AuthSchemeType,
    IntegrationModel,
)
from composio.client.exceptions import ComposioClientError
from composio.constants import DEFAULT_ENTITY_ID
from composio.utils.url import get_web_url


class AddIntegrationExamples(HelpfulCmd):
    examples = [
        click.style("composio add <app_name>", fg="green")
        + click.style("                      # Add a new integration\n", fg="black"),
        click.style("composio add <app_name> --no-browser", fg="green")
        + click.style(
            "         # Add a new integration without opening the browser\n", fg="black"
        ),
        click.style("composio add <app_name> -i <integration_id>", fg="green")
        + click.style(
            "  # Add a new integration using an existing integration ID\n", fg="black"
        ),
    ]


@click.command(name="add", cls=AddIntegrationExamples)
@click.help_option("--help", "-h", "-help")
@click.argument("name", type=str)
@click.option(
    "--no-browser",
    is_flag=True,
    default=False,
    help="Don't open browser for verifying connection",
)
@click.option(
    "-i",
    "--integration-id",
    type=str,
    help="Specify integration ID to use existing integration",
)
@click.option(
    "-a",
    "--auth-mode",
    type=str,
    help="Specify auth mode for given app.",
)
@click.option(
    "-s",
    "--scope",
    "scopes",
    type=str,
    help="Specify scopes for the connection.",
    multiple=True,
)
@click.option(
    "--force",
    is_flag=True,
    help="Override the existing account.",
)
@click.option(
    "-l",
    "--label",
    "labels",
    help="Labels for connected account.",
    multiple=True,
)
@pass_entity_id
@handle_exceptions()
@ensure_login
@pass_context
def _add(
    context: Context,
    name: str,
    scopes: t.Tuple[str, ...],
    entity_id: str,
    integration_id: t.Optional[str],
    labels: t.List[str],
    no_browser: bool = False,
    auth_mode: t.Optional[str] = None,
    force: bool = False,
) -> None:
    """Add a new integration."""
    add_integration(
        name=name.lower().strip(),
        context=context,
        entity_id=entity_id,
        integration_id=integration_id,
        no_browser=no_browser,
        auth_mode=auth_mode,
        scopes=scopes,
        force=force,
        labels=list(labels or []),
    )


def _replace_connection() -> bool:
    """Prompt user to check if they want to replace the connection or not."""
    return (
        click.prompt(
            "> Do you want to replace the existing connection?",
            type=click.Choice(
                choices=("y", "n"),
                case_sensitive=False,
            ),
        )
        == "y"
    )


def _collect_input_fields(fields: t.List[AuthSchemeField]) -> t.Dict:
    """Collect"""
    inputs = {}
    for _field in fields:
        field = _field.model_dump()
        if field.get("required", False):
            value = input(f"> Enter {field.get('display_name', field.get('name'))}: ")
            if not value:
                raise click.ClickException(
                    f"{field.get('display_name', field.get('name'))} is required"
                )
        else:
            value = input(
                f"> Enter {field.get('display_name', field.get('name'))} (Optional):"
            ) or t.cast(
                str,
                field.get("default"),
            )
        inputs[field.get("name")] = value
    return inputs


def _load_integration(
    context: Context,
    integration_id: t.Optional[str] = None,
) -> t.Optional[IntegrationModel]:
    """Load integration model."""
    if integration_id is None:
        return None

    for integration in context.client.integrations.get():
        if integration.id == integration_id:
            return integration

    raise click.ClickException(f"No integration found with ID: `{integration_id}`")


def add_integration(
    name: str,
    context: Context,
    entity_id: str = DEFAULT_ENTITY_ID,
    integration_id: t.Optional[str] = None,
    no_browser: bool = False,
    auth_mode: t.Optional[str] = None,
    scopes: t.Optional[t.Tuple[str, ...]] = None,
    labels: t.Optional[t.List] = None,
    force: bool = False,
) -> None:
    """
    Add integration.

    :param name: App name.
    :param context: CLI runtime context.
    :param entity_id: Entity ID to use for creating integration.
    :param no_browser: Don't open browser.
    :param auth_mode: Preferred auth mode.
    :param scopes: List of scopes for the connected account.
    """
    entity = context.client.get_entity(id=entity_id)
    integration = _load_integration(
        context=context,
        integration_id=integration_id,
    )

    try:
        existing_connection = entity.get_connection(app=name)
    except ComposioClientError:
        existing_connection = None

    if existing_connection is not None and not force:
        context.console.print(
            f"[yellow]Warning: An existing connection for {name} was found.[/yellow]\n"
        )
        if not _replace_connection():
            context.console.print(
                "\n[green]Existing connection retained. No new connection added.[/green]\n"
            )
            return None

    if existing_connection is not None and force:
        context.console.print(
            f"[yellow]Warning: Replacing existing connection for {name}.[/yellow]\n"
        )

    context.console.print(
        f"\n[green]> Adding integration: {name.capitalize()}...[/green]\n"
    )
    app = t.cast(AppModel, context.client.apps.get(name=name))
    if app.no_auth:
        raise click.ClickException(f"{app.name} does not require authentication")

    auth_schemes = app.auth_schemes or []
    if len(auth_schemes) == 0:
        context.console.print(f"{app.name} does not need authentication")
        return None

    auth_modes = {auth_scheme.auth_mode: auth_scheme for auth_scheme in auth_schemes}
    if auth_mode is not None and auth_mode not in auth_modes:
        raise click.ClickException(
            f"Invalid value for `auth_mode`, select from `{set(auth_modes)}`"
        )

    if auth_mode is not None:
        auth_mode = t.cast(AuthSchemeType, auth_mode)
        auth_scheme = auth_modes[auth_mode]
    elif len(auth_modes) == 1:
        ((auth_mode, auth_scheme),) = auth_modes.items()
    else:
        auth_mode = t.cast(
            AuthSchemeType,
            click.prompt(
                "Select auth mode: ",
                type=click.Choice(choices=list(auth_modes)),
            ),
        )
        auth_scheme = auth_modes[auth_mode]

    if auth_mode.lower() in ("basic", "api_key", "bearer_token"):
        return _handle_basic_auth(
            entity=entity,
            client=context.client,
            app_name=name,
            auth_mode=auth_mode,
            auth_scheme=auth_scheme,
            labels=labels,
        )
    return _handle_oauth(
        entity=entity,
        client=context.client,
        app_name=name,
        auth_scheme=auth_scheme,
        no_browser=no_browser,
        integration=integration,
        scopes=scopes,
        use_composio_auth=len(app.testConnectors or []) != 0,
        labels=labels,
    )


def _handle_oauth(
    entity: Entity,
    client: Composio,
    app_name: str,
    auth_scheme: AppAuthScheme,
    no_browser: bool = False,
    integration: t.Optional[IntegrationModel] = None,
    scopes: t.Optional[t.Tuple[str, ...]] = None,
    labels: t.Optional[t.List] = None,
    use_composio_auth: bool = False,
) -> None:
    """Handle no auth."""
    auth_config = {}
    if not use_composio_auth:
        auth_config.update(
            _collect_input_fields(
                fields=auth_scheme.fields,
            )
        )

    if scopes is not None:
        if auth_config.get("scopes") is not None:
            scopes = tuple(set([*auth_config["scopes"].split(","), *scopes]))
        auth_config["scopes"] = ",".join(scopes)

    connection = entity.initiate_connection(
        app_name=app_name.lower(),
        redirect_url=get_web_url(path="redirect"),
        integration=integration,
        auth_mode="OAUTH2",
        auth_config=auth_config,
        use_composio_auth=use_composio_auth,
        force_new_integration=len(scopes or []) > 0,
        labels=labels,
    )
    if not no_browser:
        webbrowser.open(
            url=str(connection.redirectUrl),
        )
    click.echo(
        f"Please authenticate {app_name} in the browser and come back here. "
        f"URL: {connection.redirectUrl}"
    )
    click.echo(f"⚠ Waiting for {app_name} authentication...")
    connection.wait_until_active(client=client)
    click.echo(
        f"✔ {app_name} added successfully with ID: {connection.connectedAccountId}"
    )


def _handle_basic_auth(
    entity: Entity,
    client: Composio,
    app_name: str,
    auth_mode: str,
    auth_scheme: AppAuthScheme,
    integration: t.Optional[IntegrationModel] = None,
    labels: t.Optional[t.List] = None,
) -> None:
    """Handle basic auth."""
    auth_config = _collect_input_fields(
        fields=auth_scheme.fields,
    )
    connection = entity.initiate_connection(
        app_name=app_name.lower(),
        auth_mode=auth_mode,
        auth_config=auth_config,
        integration=integration,
        use_composio_auth=False,
        force_new_integration=True,
        labels=labels,
    )
    connection.save_user_access_data(
        client=client,
        field_inputs=auth_config,
        entity_id=entity.id,
    )
    click.echo(
        f"✔ {app_name} added successfully with ID: {connection.connectedAccountId}"
    )



"""
Zep as composio tool.

Read about zep - https://help.getzep.com/concepts
"""

import typing as t

from composio.tools.base.local import LocalAction, LocalTool
from composio.tools.local.zep.actions.add_memory import AddMemory
from composio.tools.local.zep.actions.create_session import CreateSession
from composio.tools.local.zep.actions.get_memory import GetMemory
from composio.tools.local.zep.actions.search_memory import SearchMemory


class Zeptool(LocalTool, autoload=True):
    """Tool definition for zep"""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/zep.png"

    @classmethod
    def actions(cls) -> t.List[t.Type[LocalAction]]:
        """Get zep actions."""
        return [
            CreateSession,
            AddMemory,
            GetMemory,
            SearchMemory,
        ]



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateTaskFromTemplateRequest(BaseModel):
    """Request schema for `CreateTaskFromTemplate`"""

    list_id: int = Field(
        ...,
        alias="list_id",
        description="",
    )
    template_id: str = Field(
        ...,
        alias="template_id",
        description="",
    )
    name: str = Field(
        default=...,
        alias="name",
        description="Name",
    )


class CreateTaskFromTemplateResponse(BaseModel):
    """Response schema for `CreateTaskFromTemplate`"""

    data: t.Dict[str, t.Any]


class CreateTaskFromTemplate(OpenAPIAction):
    """Create a new task using a task template."""

    _tags = ["Task Templates"]
    _display_name = "create_task_from_template"
    _request_schema = CreateTaskFromTemplateRequest
    _response_schema = CreateTaskFromTemplateResponse

    url = "https://api.clickup.com/api/v2"
    path = "/list/{list_id}/taskTemplate/{template_id}"
    method = "post"
    operation_id = "TaskTemplates_createFromTemplate"
    action_identifier = "/list/{list_id}/taskTemplate/{template_id}_post"

    path_params = {"list_id": "list_id", "template_id": "template_id"}
    query_params = {}
    header_params = {}
    request_params = {"name": {"__alias": "name"}}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetSpaceRequest(BaseModel):
    """Request schema for `GetSpace`"""

    space_id: int = Field(
        ...,
        alias="space_id",
        description="",
    )


class GetSpaceResponse(BaseModel):
    """Response schema for `GetSpace`"""

    data: t.Dict[str, t.Any]


class GetSpace(OpenAPIAction):
    """View the Spaces available in a Workspace."""

    _tags = ["Spaces"]
    _display_name = "get_space"
    _request_schema = GetSpaceRequest
    _response_schema = GetSpaceResponse

    url = "https://api.clickup.com/api/v2"
    path = "/space/{space_id}"
    method = "get"
    operation_id = "Spaces_getDetails"
    action_identifier = "/space/{space_id}_get"

    path_params = {"space_id": "space_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}


class spaces_get_details(OpenAPIAction):
    """View the Spaces available in a Workspace.<<DEPRECATED use get_space>>"""

    _tags = ["Spaces"]
    _display_name = "spaces_get_details"
    _request_schema = GetSpaceRequest
    _response_schema = GetSpaceResponse

    url = "https://api.clickup.com/api/v2"
    path = "/space/{space_id}"
    method = "get"
    operation_id = "Spaces_getDetails"
    action_identifier = "/space/{space_id}_get"

    path_params = {"space_id": "space_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetTaskCommentsRequest(BaseModel):
    """Request schema for `GetTaskComments`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )
    start: int = Field(
        default=None,
        alias="start",
        description="Enter the `date` of a task comment using Unix time in milliseconds.",
    )
    start_id: str = Field(
        default=None,
        alias="start_id",
        description="Enter the Comment `id` of a task comment.",
    )


class GetTaskCommentsResponse(BaseModel):
    """Response schema for `GetTaskComments`"""

    data: t.Dict[str, t.Any]


class GetTaskComments(OpenAPIAction):
    """
    View task comments.    If you do not include the `start` and `start_id`
    parameters, this endpoint will return the most recent 25 comments.   Use
    the `start` and `start id` parameters of the oldest comment to retrieve
    the next 25 comments.
    """

    _tags = ["Comments"]
    _display_name = "get_task_comments"
    _request_schema = GetTaskCommentsRequest
    _response_schema = GetTaskCommentsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/comment"
    method = "get"
    operation_id = "Comments_getTaskComments"
    action_identifier = "/task/{task_id}/comment_get"

    path_params = {"task_id": "task_id"}
    query_params = {
        "custom_task_ids": "custom_task_ids",
        "team_id": "team_id",
        "start": "start",
        "start_id": "start_id",
    }
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class UpdateTaskRequest(BaseModel):
    """Request schema for `UpdateTask`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )
    description: str = Field(
        default=None,
        alias="description",
        description='To clear the task description, include `Description` with `" "`.',
    )
    custom_item_id: int = Field(
        default=None,
        alias="custom_item_id",
        description=(
            'To convert an item using a custom task type into a task, send `"null"`. '
            "   To update this task to be a Milestone, send a value of `1`.    To use "
            "a custom task type, send the custom task type ID as defined in your Workspace, "
            "such as `2`. "
        ),
    )
    name: str = Field(
        default=None,
        alias="name",
        description="Name",
    )
    status: str = Field(
        default=None,
        alias="status",
        description="Status",
    )
    priority: int = Field(
        default=None,
        alias="priority",
        description="Priority",
    )
    due_date: int = Field(
        default=None,
        alias="due_date",
        description="Due Date",
    )
    due_date_time: bool = Field(
        default=None,
        alias="due_date_time",
        description="Due Date Time",
    )
    parent: str = Field(
        default=None,
        alias="parent",
        description=(
            'You can move a subtask to another parent task by including `"parent"` with '
            'a valid `task id`.   You cannot convert a subtask to a task by setting `"parent"` '
            "to `null`. "
        ),
    )
    time_estimate: int = Field(
        default=None,
        alias="time_estimate",
        description="Time Estimate",
    )
    start_date: int = Field(
        default=None,
        alias="start_date",
        description="Start Date",
    )
    start_date_time: bool = Field(
        default=None,
        alias="start_date_time",
        description="Start Date Time",
    )
    assignees_add: t.List[int] = Field(
        default=...,
        alias="assignees__add",
        description="",
    )
    assignees_rem: t.List[int] = Field(
        default=...,
        alias="assignees__rem",
        description="",
    )
    archived: bool = Field(
        default=None,
        alias="archived",
        description="Archived",
    )


class UpdateTaskResponse(BaseModel):
    """Response schema for `UpdateTask`"""

    data: t.Dict[str, t.Any]


class UpdateTask(OpenAPIAction):
    """Update a task by including one or more fields in the request body."""

    _tags = ["Tasks"]
    _display_name = "update_task"
    _request_schema = UpdateTaskRequest
    _response_schema = UpdateTaskResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}"
    method = "put"
    operation_id = "Tasks_updateTaskFields"
    action_identifier = "/task/{task_id}_put"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {
        "description": {"__alias": "description"},
        "custom_item_id": {"__alias": "custom_item_id"},
        "name": {"__alias": "name"},
        "status": {"__alias": "status"},
        "priority": {"__alias": "priority"},
        "due_date": {"__alias": "due_date"},
        "due_date_time": {"__alias": "due_date_time"},
        "parent": {"__alias": "parent"},
        "time_estimate": {"__alias": "time_estimate"},
        "start_date": {"__alias": "start_date"},
        "start_date_time": {"__alias": "start_date_time"},
        "assignees": {
            "__alias": "assignees",
            "add": {"__alias": "add"},
            "rem": {"__alias": "rem"},
        },
        "archived": {"__alias": "archived"},
    }

    aliases = {"assignees": "24afd83d"}



"""
Logging utilities.
"""

import atexit
import logging
import os
import threading
import typing as t
from enum import Enum
from queue import Empty, Queue

from composio.constants import ENV_COMPOSIO_LOGGING_LEVEL
from composio.exceptions import ApiKeyNotProvidedError


_DEFAULT_FORMAT = "[%(asctime)s][%(levelname)s] %(message)s"
_DEFAULT_LOGGER_NAME = "composio"

_LOG_VERBOSITY = int(os.environ.get("COMPOSIO_LOG_VERBOSITY", 0))
_LOG_LINE_SIZE_BY_VERBOSITY = {
    0: 256,
    1: 512,
    2: 1024,
    3: -1,
}

_logger: t.Optional[logging.Logger] = None
"""Global logger object for composio."""

_logger_wrapper: t.Optional["_VerbosityWrapper"] = None
"""Global logger object wrapped with verbosity setting for composio."""


class LogLevel(Enum):
    """Logging level."""

    CRITICAL = "critical"
    FATAL = "fatal"
    ERROR = "error"
    WARNING = "warning"
    WARN = "warn"
    INFO = "info"
    DEBUG = "debug"
    NOTSET = "notset"


_LEVELS: t.Dict[LogLevel, int] = {
    LogLevel.CRITICAL: logging.CRITICAL,
    LogLevel.FATAL: logging.FATAL,
    LogLevel.ERROR: logging.ERROR,
    LogLevel.WARNING: logging.WARNING,
    LogLevel.WARN: logging.WARN,
    LogLevel.INFO: logging.INFO,
    LogLevel.DEBUG: logging.DEBUG,
    LogLevel.NOTSET: logging.NOTSET,
}


class _VerbosityWrapper:
    def __init__(
        self,
        logger: logging.Logger,
        verbosity_level: t.Optional[int] = None,
    ) -> None:
        self.logger = logger
        self.level = logger.level
        self.setup(verbosity_level=verbosity_level)
        self.verbosity = min(verbosity_level or _LOG_VERBOSITY, 3)
        self.size = _LOG_LINE_SIZE_BY_VERBOSITY[self.verbosity]

    def setup(self, verbosity_level: t.Optional[int] = None) -> None:
        if verbosity_level is None:
            return
        self.verbosity = verbosity_level
        self.size = _LOG_LINE_SIZE_BY_VERBOSITY[self.verbosity]

    def _trim(self, msg) -> str:
        msg = str(msg)
        if self.size == -1:
            return msg

        if len(msg) < self.size:
            return msg

        return msg[: self.size] + "..."

    def info(self, msg, *args, **kwargs):
        self.logger.info(self._trim(msg), *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(self._trim(msg), *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(self._trim(msg), *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)


def _parse_log_level_from_env(default: int) -> int:
    """Parse log level from environment."""
    level = os.environ.get(ENV_COMPOSIO_LOGGING_LEVEL)
    if level is None:
        return default

    try:
        return _LEVELS[LogLevel(level.lower())]
    except (ValueError, KeyError):
        return default


def setup(level: LogLevel = LogLevel.INFO, log_format: str = _DEFAULT_FORMAT) -> None:
    """Setup logging config."""
    global _logger_wrapper
    if _logger_wrapper is None:
        _logger_wrapper = get(name=_DEFAULT_LOGGER_NAME)

    logging.basicConfig(format=log_format)
    _logger_wrapper.logger.setLevel(_LEVELS[level])


def get(
    name: t.Optional[str] = None,
    level: LogLevel = LogLevel.INFO,
    log_format: str = _DEFAULT_FORMAT,
    verbosity_level: t.Optional[int] = None,
) -> _VerbosityWrapper:
    """Set up the logger."""
    global _logger, _logger_wrapper
    if _logger_wrapper is not None:
        return t.cast(_VerbosityWrapper, _logger_wrapper)

    # Setup logging format.
    logging.basicConfig(format=log_format)

    # Create logger
    _logger = logging.getLogger(name or _DEFAULT_LOGGER_NAME)
    _logger.setLevel(_parse_log_level_from_env(default=_LEVELS[level]))
    _logger_wrapper = _VerbosityWrapper(_logger, verbosity_level=verbosity_level)
    return _logger_wrapper


class WithLogger:
    """Interface to endow subclasses with a logger."""

    def __init__(
        self,
        logger: t.Optional[logging.Logger] = None,
        logger_name: str = _DEFAULT_LOGGER_NAME,
        logging_level: LogLevel = LogLevel.INFO,
        verbosity_level: t.Optional[int] = None,
    ) -> None:
        """
        Initialize the logger.

        :param logger: the logger object.
        :param logger_name: the default logger name, if a logger is not provided.
        """
        self._logger = (
            _VerbosityWrapper(logger, verbosity_level=verbosity_level)
            if logger is not None
            else get(name=logger_name, level=logging_level)
        )
        self._logger.setup(verbosity_level=verbosity_level)
        self._logging_level = logging._levelToName[self._logger.level]

    @property
    def logger(self) -> logging.Logger:
        """Get the component logger."""
        return t.cast(logging.Logger, self._logger)


class LogIngester:
    def __init__(self) -> None:
        from composio import Composio  # pylint: disable=import-outside-toplevel

        try:
            self._client = Composio.get_latest()
        except ApiKeyNotProvidedError:
            return

        self._queue = Queue[t.Dict]()
        self._event = threading.Event()
        self._thread = threading.Thread(target=self._wait, daemon=True)
        self._thread.start()

        @atexit.register
        def _teardown():
            self._teardown()

    def log(
        self,
        connection_id,
        provider_name,
        action_name,
        request,
        response,
        is_error,
        session_id,
    ) -> None:
        """Log new request."""
        if not hasattr(self, "_queue"):
            return

        self._queue.put_nowait(
            item={
                "connectionId": connection_id,
                "providerName": provider_name,
                "actionName": action_name,
                "request": request,
                "response": response,
                "isError": is_error,
                "sessionId": session_id,
            }
        )

    def _push(self, record: t.Dict) -> None:
        """Push log to server."""
        self._client.logs.push(record=record)

    def _wait(self) -> None:
        """Wait for the logs."""
        while not self._event.is_set():
            try:
                self._push(record=self._queue.get(timeout=0.5))
            except Empty:
                continue

    def _teardown(self) -> None:
        """Teardown the logging client."""
        self._event.set()
        self._thread.join()


get_logger = get



"""
Docker workspace.
"""

import os
import socket
import time
import typing as t
from dataclasses import dataclass
from pathlib import Path

from composio.exceptions import ComposioSDKError
from composio.tools.env.base import RemoteWorkspace, WorkspaceConfigType
from composio.tools.env.constants import (
    DEFAULT_IMAGE,
    ENV_COMPOSIO_DEV_MODE,
    ENV_COMPOSIO_TOOLSERVER_IMAGE,
)


try:
    import requests
    from docker import DockerClient, from_env
    from docker.errors import DockerException, NotFound
    from docker.models.containers import Container

    DOCKER_INSTALLED = True
except ImportError:
    DockerClient = t.Any
    Container = t.Any

    NotFound = Exception
    DockerException = Exception
    DOCKER_INSTALLED = False


COMPOSIO_PATH = Path(__file__).parent.parent.parent.parent.parent.resolve()
COMPOSIO_CACHE = Path.home() / ".composio"
CONTAINER_DEV_VOLUMES = {
    COMPOSIO_PATH: {
        "bind": "/opt/composio-core",
        "mode": "rw",
    },
    COMPOSIO_CACHE: {
        "bind": "/root/.composio",
        "mode": "rw",
    },
}
DEV_MODE = os.environ.get(ENV_COMPOSIO_DEV_MODE, "0") == "1"
DEFAULT_PORT = 54321


def _get_free_port() -> int:
    sock = socket.socket()
    try:
        sock.bind(("", 0))
        return sock.getsockname()[1]
    finally:
        sock.close()


def _kill_docker_container(container: str, client: DockerClient):
    try:
        client.api.kill(container)
    except DockerException:
        pass

    try:
        client.api.remove_container(container)
    except DockerException:
        pass


@dataclass
class Config(WorkspaceConfigType):
    """Docker configuration type."""

    image: t.Optional[str] = None
    """Name of the docker image."""

    ports: t.Optional[t.Dict[int, t.Any]] = None
    """
    Ports to bind inside the container

    Note: port 8000 is reserved for the tooling server inside the container
    """

    volumes: t.Optional[t.Dict[str, t.Any]] = None
    """Voluems to bind inside the container"""


class DockerWorkspace(RemoteWorkspace):
    """Docker workspace implementation."""

    _port: int  # Tooling server port
    _container: Container
    _client: t.Optional[DockerClient] = None

    def __init__(self, config: Config) -> None:
        """Create a docker workspace."""
        if not DOCKER_INSTALLED:
            raise ComposioSDKError(
                "`docker` is required to use docker workspace, "
                "run `pip3 install composio-core[docker]` or "
                "`pip3 install docker` to install"
            )

        super().__init__(config=config)
        self.image = config.image or os.environ.get(
            ENV_COMPOSIO_TOOLSERVER_IMAGE,
            DEFAULT_IMAGE,
        )
        self._port_requests = config.ports or {}
        self._volume_requests = config.volumes or {}

    def setup(self) -> None:
        """Setup docker workspace."""
        self.logger.debug(f"Creating docker workspace with image: {self.image}")
        self._port = _get_free_port()

        container_kwargs: t.Dict[str, t.Any] = {
            "tty": True,
            "detach": True,
            "stdin_open": True,
            "auto_remove": False,
            "image": self.image,
            "name": self.id,
            "environment": self.environment,
            "command": "/root/entrypoint.sh",
            "ports": {
                **self._port_requests,
                8000: self._port,
            },
            "volumes": self._volume_requests,
        }
        if DEV_MODE:
            container_kwargs["volumes"].update(CONTAINER_DEV_VOLUMES)  # type: ignore
            container_kwargs["environment"][ENV_COMPOSIO_DEV_MODE] = "1"

        while True:
            try:
                self._container = self.client.containers.run(**container_kwargs)
                self._container.start()
                self._container.reload()
                break
            except DockerException as e:
                if "Ports are not available" not in str(e):
                    raise ComposioSDKError("Error starting workspace: ", e) from e

                _kill_docker_container(container=self.id, client=self.client)
                self._port = _get_free_port()
                container_kwargs["ports"][8000] = self._port

        # Wait for tooling server to start
        self.url = f"http://localhost:{self._port}/api"
        self._wait()

        # Setup network config
        self.host = "localhost"
        self.ports = [
            int(port[0]["HostPort"])
            for port in self._container.ports.values()
            if int(port[0]["HostPort"]) != self._port
        ]

    def _wait(self) -> None:
        """Wait for docker workspace to get started."""
        deadline = time.time() + float(os.environ.get("WORKSPACE_WAIT_TIMEOUT", 60.0))
        while time.time() < deadline:
            try:
                if (
                    self._request(endpoint="", method="get", log=False).status_code
                    == 200
                ):
                    return
            except requests.ConnectionError:
                time.sleep(1)

        self.logger.error(
            "Timed out while waiting for docker workspace to start\n"
            f"{self._container.logs().decode(encoding='utf-8')}"
        )
        raise ComposioSDKError(
            message="Timed out while waiting for docker workspace to start"
        )

    @property
    def client(self) -> DockerClient:
        """Docker client object."""
        if self._client is None:
            try:
                self._client = from_env(timeout=100)
            except DockerException as e:
                raise ComposioSDKError(
                    message=f"Error initializing docker client: {e}. "
                    "Please make sure docker is running and try again."
                ) from e
        return self._client

    def teardown(self) -> None:
        """Teardown docker workspace factory."""
        super().teardown()
        try:
            self._container.kill()
            self._container.remove()
        except NotFound as e:
            self.logger.debug(f"Error cleaning {self.id} - {e}")



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetGoalsRequest(BaseModel):
    """Request schema for `GetGoals`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    include_completed: bool = Field(
        default=None,
        alias="include_completed",
        description="",
    )


class GetGoalsResponse(BaseModel):
    """Response schema for `GetGoals`"""

    data: t.Dict[str, t.Any]


class GetGoals(OpenAPIAction):
    """View the Goals available in a Workspace."""

    _tags = ["Goals"]
    _display_name = "get_goals"
    _request_schema = GetGoalsRequest
    _response_schema = GetGoalsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/goal"
    method = "get"
    operation_id = "Goals_getWorkspaceGoals"
    action_identifier = "/team/{team_id}/goal_get"

    path_params = {"team_id": "team_id"}
    query_params = {"include_completed": "include_completed"}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetListRequest(BaseModel):
    """Request schema for `GetList`"""

    list_id: int = Field(
        ...,
        alias="list_id",
        description="",
    )


class GetListResponse(BaseModel):
    """Response schema for `GetList`"""

    data: t.Dict[str, t.Any]


class GetList(OpenAPIAction):
    """View information about a List."""

    _tags = ["Lists"]
    _display_name = "get_list"
    _request_schema = GetListRequest
    _response_schema = GetListResponse

    url = "https://api.clickup.com/api/v2"
    path = "/list/{list_id}"
    method = "get"
    operation_id = "Lists_getListDetails"
    action_identifier = "/list/{list_id}_get"

    path_params = {"list_id": "list_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import Field

from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class FindFileRequest(BaseFileRequest):
    """Request to find files matching a pattern."""

    pattern: str = Field(
        ...,
        description="Pattern to search for (supports wildcards)",
    )
    depth: t.Optional[int] = Field(
        default=None,
        description="Max depth to search for (None for unlimited)",
        ge=0,
    )
    case_sensitive: bool = Field(
        default=False,
        description="If set True the search will be case sensitive",
    )
    include: t.List[str] = Field(
        default=None,
        description="List of directories to search in",
    )
    exclude: t.List[str] = Field(
        default=None,
        description="List of directories to exclude from the search",
    )


class FindFileResponse(BaseFileResponse):
    """Response to find files matching a pattern."""

    results: t.List[str] = Field(
        default=[],
        description="List of file paths matching the search pattern",
    )
    message: str = Field(
        default="",
        description="Message to display to the user",
    )
    error: str = Field(
        default="",
        description="Error message if any",
    )


class FindFile(LocalAction[FindFileRequest, FindFileResponse]):
    """
    Finds files or directories matching the given pattern in the workspace.

    This action allows you to search for files or directories using glob patterns.
    You can specify various parameters to refine your search, such as search depth,
    case sensitivity, directories to include or exclude.

    The search returns a list of file paths relative to the working directory that
    match the given pattern.

    Usage examples:
    1. All Python files: "*.py"
    2. Text files starting with "test_": "test_*.txt"
    3. Markdown files in any subdirectory: "**/*.md"
    4. CSV files like "data001.csv", "data002.csv": "data???.csv"
    5. "main.js" in "src" and subdirs: "src/**/main.js"

    Note: The search automatically excludes the '.git' directory.

    Returns:
    - A list of file paths relative to the working directory that match the search pattern.
    """

    @include_cwd  # type: ignore
    def execute(self, request: FindFileRequest, metadata: t.Dict) -> FindFileResponse:
        try:
            results = self.filemanagers.get(request.file_manager_id).find(
                pattern=request.pattern,
                depth=request.depth,
                case_sensitive=request.case_sensitive,
                include=request.include,  # type: ignore
                exclude=request.exclude,  # type: ignore
            )
            if len(results) > 200:
                return FindFileResponse(
                    results=results[:200],
                    message=(
                        f"Too many results found. Found {len(results)} results, "
                        "returning 300 of them. Please refine your search criteria."
                    ),
                )
            if results == []:
                return FindFileResponse(error="No results found.")
            return FindFileResponse(results=results)
        except ValueError as e:
            return FindFileResponse(error=f"Invalid search parameters: {str(e)}")
        except PermissionError as e:
            return FindFileResponse(error=f"Permission denied: {str(e)}")
        except OSError as e:
            return FindFileResponse(error=f"File system error: {str(e)}")



"""
Action for executing a custom JavaScript script on a webpage.
"""

from typing import Any, Optional

from pydantic import Field

from composio.tools.env.browsermanager.manager import BrowserManager
from composio.tools.local.browsertool.actions.base_action import (
    BaseBrowserAction,
    BaseBrowserRequest,
    BaseBrowserResponse,
)


class ExecuteScriptRequest(BaseBrowserRequest):
    """Request schema for executing a custom JavaScript script."""

    script: str = Field(..., description="JavaScript code to be executed")
    args: Optional[list] = Field(
        default=None, description="Optional arguments to pass to the script"
    )


class ExecuteScriptResponse(BaseBrowserResponse):
    """Response schema for executing a custom JavaScript script."""

    success: bool = Field(
        default=False, description="Whether the script execution was successful"
    )
    result: Optional[Any] = Field(
        default=None, description="The result of the script execution"
    )


class ExecuteScript(BaseBrowserAction[ExecuteScriptRequest, ExecuteScriptResponse]):
    """
    Execute a custom JavaScript script on the current webpage.

    This action allows you to run arbitrary JavaScript code in the context of the current webpage.

    It can be used for various purposes such as:
    - Extracting data from the page that's not easily accessible through standard selectors
    - Modifying the page content or structure

    Example could be
    () => {
      return document.title;
    }
    """

    display_name = "ExecuteScript"
    _request_schema = ExecuteScriptRequest
    _response_schema = ExecuteScriptResponse

    def execute_on_browser_manager(
        self, browser_manager: BrowserManager, request: ExecuteScriptRequest  # type: ignore
    ) -> ExecuteScriptResponse:
        """Execute the custom JavaScript script."""

        try:
            # Execute the script
            result = browser_manager.execute_script(
                request.script, *(request.args or [])
            )

            return ExecuteScriptResponse(success=True, result=result)
        except Exception as e:
            return ExecuteScriptResponse(
                success=False, error=f"Failed to execute script: {str(e)}"
            )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateWebhookRequest(BaseModel):
    """Request schema for `CreateWebhook`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    endpoint: str = Field(
        default=...,
        alias="endpoint",
        description="Endpoint",
    )
    events: t.List[str] = Field(
        default=...,
        alias="events",
        description="Events",
    )
    space_id: int = Field(
        default=None,
        alias="space_id",
        description="Space Id",
    )
    folder_id: int = Field(
        default=None,
        alias="folder_id",
        description="Folder Id",
    )
    list_id: int = Field(
        default=None,
        alias="list_id",
        description="List Id",
    )
    task_id: str = Field(
        default=None,
        alias="task_id",
        description="Task Id",
    )


class CreateWebhookResponse(BaseModel):
    """Response schema for `CreateWebhook`"""

    data: t.Dict[str, t.Any]


class CreateWebhook(OpenAPIAction):
    """
    Set up a webhook to monitor for events.<br> We do not have a dedicated IP
    address for webhooks. We use our domain name and dynamic addressing.
    """

    _tags = ["Webhooks"]
    _display_name = "create_webhook"
    _request_schema = CreateWebhookRequest
    _response_schema = CreateWebhookResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/webhook"
    method = "post"
    operation_id = "Webhooks_createWebhook"
    action_identifier = "/team/{team_id}/webhook_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "endpoint": {"__alias": "endpoint"},
        "events": {"__alias": "events"},
        "space_id": {"__alias": "space_id"},
        "folder_id": {"__alias": "folder_id"},
        "list_id": {"__alias": "list_id"},
        "task_id": {"__alias": "task_id"},
    }

    aliases = {}



"""
Tool schema helpers
"""

import typing as t
from enum import Enum

from pydantic import BaseModel


class OpenAISchema(BaseModel):
    type: str
    function: t.Dict[str, t.Any]


class ClaudeSchema(BaseModel):
    name: str
    description: str
    input_schema: t.Dict[str, t.Any]


class SchemaType(Enum):
    """Schema type."""

    CLAUDE = "claude"
    OPENAI = "openai"
    DEFAULT = "default"

    def format(
        self,
        schema: t.Dict,
    ) -> t.Union[t.Dict[str, t.Any], OpenAISchema, ClaudeSchema]:
        """Format action schema."""
        if self == SchemaType.OPENAI:
            return OpenAISchema(
                type="function",
                function={
                    "name": schema["name"],
                    "description": schema.get("description", ""),
                    "parameters": schema.get("parameters", {}),
                },
            )

        if self == SchemaType.CLAUDE:
            return ClaudeSchema(
                name=schema["name"],
                description=schema.get("description", ""),
                input_schema=schema.get("parameters", {}),
            )

        return schema



"""
API Endpoints.
"""

# pylint: disable=consider-using-with, subprocess-run-check, unspecified-encoding

import importlib
import os
import subprocess
import sys
import tempfile
import traceback
import typing as t
import zipfile
from functools import update_wrapper
from hashlib import md5
from pathlib import Path

import typing_extensions as te
from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.responses import FileResponse
from pydantic import BaseModel, Field

from composio import Action, App
from composio.cli.context import get_context
from composio.client.collections import ActionModel, AppModel
from composio.client.enums.base import get_runtime_actions
from composio.tools.base.abs import action_registry
from composio.tools.env.base import ENV_ACCESS_TOKEN
from composio.tools.local import load_local_tools
from composio.utils.logging import get as get_logger


ResponseType = t.TypeVar("ResponseType")

R = t.TypeVar("R")
T = t.TypeVar("T")
P = te.ParamSpec("P")
F = t.TypeVar("F")


class APIResponse(BaseModel, t.Generic[ResponseType]):
    """API Response."""

    data: t.Optional[ResponseType]
    error: t.Optional[str] = None
    traceback: t.Optional[str] = None


class GetApiResponse(BaseModel):
    """Response for GET /api."""

    version: str = Field(
        ...,
        description="Current API version.",
    )


class ToolUploadRequest(BaseModel):
    """Tool upload request."""

    content: str = Field(
        ...,
        description="Content from the tool description file.",
    )
    filename: str = Field(
        ...,
        description="Name of the file.",
    )
    dependencies: t.List[str] = Field(
        ...,
        description="List of dependencies.",
    )


class ExecuteActionRequest(BaseModel):
    """Execute action request."""

    params: t.Dict = Field(
        ...,
        description="Parameters for executing the request.",
    )
    metadata: t.Dict = Field(
        None,
        description="Metadata for executing action.",
    )
    entity_id: str = Field(
        None,
        description="Entity ID associated with the account.",
    )
    connected_account_id: str = Field(
        None,
        description="Connection ID to use for executing the action.",
    )


class ValidateToolsRequest(BaseModel):
    apps: t.Optional[t.List[str]] = Field(
        None,
        description="Apps list.",
    )
    actions: t.Optional[t.List[str]] = Field(
        None,
        description="Actions list.",
    )
    tags: t.Optional[t.List[str]] = Field(
        None,
        description="Tags list.",
    )


def create_app() -> FastAPI:
    """Create Fast API app."""
    load_local_tools()

    access_token = os.environ.get(ENV_ACCESS_TOKEN)
    tooldir = tempfile.TemporaryDirectory()
    app = FastAPI(on_shutdown=[tooldir.cleanup])
    sys.path.append(tooldir.name)
    logger = get_logger()

    def with_exception_handling(f: t.Callable[P, R]) -> t.Callable[P, APIResponse[R]]:
        """Marks a callback as wanting to receive the current context object as first argument."""

        def wrapper(*args: P.args, **kwargs: P.kwargs) -> APIResponse[R]:
            try:
                return APIResponse[R](data=f(*args, **kwargs))
            except Exception as e:
                logger.error(traceback.format_exc())
                return APIResponse[R](
                    data=None,
                    error=str(e),
                    traceback=traceback.format_exc(),
                )

        return update_wrapper(wrapper, f)

    @app.middleware("http")
    async def add_process_time_header(request: Request, call_next):
        if access_token is None:
            return await call_next(request)

        if "x-api-key" in request.headers and request.headers["x-api-key"]:
            return await call_next(request)

        return Response(
            content=APIResponse[None](
                data=None,
                error="Unauthorised request",
            ).model_dump_json(),
            status_code=401,
        )

    @app.get("/api", response_model=APIResponse[GetApiResponse])
    @with_exception_handling
    def _api() -> GetApiResponse:
        """Composio tooling server API root."""
        return GetApiResponse(
            version="0.3.20",
        )

    @app.get("/api/apps", response_model=APIResponse[t.List[AppModel]])
    @with_exception_handling
    def _get_apps() -> t.List[AppModel]:
        """Get list of all available apps."""
        return get_context().client.apps.get()

    @app.post("/api/apps/update", response_model=APIResponse[bool])
    @with_exception_handling
    def _update_apps() -> bool:
        """Get list of all available apps."""
        from composio.cli.apps import (  # pylint: disable=import-outside-toplevel
            update_actions,
            update_apps,
            update_triggers,
        )

        apps = update_apps(client=get_context().client)
        update_actions(client=get_context().client, apps=apps)
        update_triggers(client=get_context().client, apps=apps)
        return True

    @app.get("/api/apps/{name}", response_model=APIResponse[AppModel])
    @with_exception_handling
    def _get_apps_by_name(name: str) -> AppModel:
        """Get list of all available apps."""
        return get_context().client.apps.get(name=name)

    @app.get("/api/actions", response_model=APIResponse[t.List[ActionModel]])
    @with_exception_handling
    def _get_actions() -> t.List[ActionModel]:
        """Get list of all available actions."""
        return get_context().client.actions.get()

    @app.get("/api/actions/{name}", response_model=APIResponse[ActionModel])
    @with_exception_handling
    def _get_actions_by_name(name: str) -> ActionModel:
        """Get list of all available apps."""
        return get_context().client.actions.get(actions=[name])[0]

    @app.get("/api/local_actions", response_model=APIResponse[t.List[ActionModel]])
    @with_exception_handling
    def _get_local_actions() -> t.List[ActionModel]:
        """Get list of all available actions."""
        return get_context().toolset.get_action_schemas(
            actions=list(action_registry["local"])
        )

    @app.get("/api/enums/actions", response_model=APIResponse[t.List[str]])
    @with_exception_handling
    def _get_actions_enums() -> t.List[str]:
        """Get list of all available actions."""
        return [action.slug for action in Action.all()]

    @app.get("/api/enums/apps", response_model=APIResponse[t.List[str]])
    @with_exception_handling
    def _get_app_enums() -> t.List[str]:
        """Get list of all available actions."""
        return [app.slug for app in App.all()]

    @app.post("/api/actions/execute/{action}", response_model=APIResponse[t.Dict])
    @with_exception_handling
    def _execute_action(action: str, request: ExecuteActionRequest) -> t.Dict:
        """Execute an action."""
        return get_context().toolset.execute_action(
            action=action,
            params=request.params,
            metadata=request.metadata,
            entity_id=request.entity_id,
            connected_account_id=request.connected_account_id,
        )

    @app.post(path="/api/validate", response_model=APIResponse[t.Dict])
    @with_exception_handling
    def _validate_tools(request: ValidateToolsRequest) -> t.Dict:
        get_context().toolset.validate_tools(
            apps=request.apps,
            actions=request.actions,
            tags=request.tags,
        )
        return {"message": "validated"}

    @app.get("/api/workspace", response_model=APIResponse[t.Dict])
    @with_exception_handling
    def _get_workspace_information() -> t.Dict:
        """Get information on current workspace."""
        return {"type": get_context().toolset.workspace.__class__.__name__}

    @app.get("/api/tools", response_model=APIResponse[t.List[str]])
    @with_exception_handling
    def _get_workspace_tools() -> t.List[str]:
        """Get list of available developer tools."""
        return get_runtime_actions()

    @app.post("/api/tools", response_model=APIResponse[t.List[str]])
    @with_exception_handling
    def _upload_workspace_tools(request: ToolUploadRequest) -> t.List[str]:
        """Get list of available developer tools."""
        if len(request.dependencies) > 0:
            process = subprocess.run(
                args=["pip", "install", *request.dependencies],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if process.returncode != 0:
                raise RuntimeError(
                    f"Error installing dependencies: {process.stderr.decode()}"
                )

        filename = md5(request.content.encode(encoding="utf-8")).hexdigest()
        tempfile = Path(tooldir.name, f"{filename}.py")
        if tempfile.exists():
            raise ValueError("Tools from this module already exits!")

        tempfile.write_text(request.content)
        importlib.import_module(filename)
        return get_runtime_actions()

    @app.get("/api/download")
    def _download_file_or_dir(file: t.Optional[str] = None):
        """Get list of available developer tools."""
        if not file:
            raise HTTPException(
                status_code=400, detail="File path is required as query parameter"
            )
        path = Path(file)
        if not path.exists():
            return Response(
                content=APIResponse[None](
                    data=None,
                    error=f"{path} not found",
                ).model_dump_json(),
                status_code=404,
            )

        if path.is_file():
            return FileResponse(path=path)

        tempdir = tempfile.TemporaryDirectory()
        zipfile = Path(tempdir.name, path.name + ".zip")
        return FileResponse(path=_archive(directory=path, output=zipfile))

    return app


def _archive(directory: Path, output: Path) -> Path:
    with zipfile.ZipFile(output, "w", zipfile.ZIP_DEFLATED) as fp:
        for root, _, files in os.walk(directory):
            for file in files:
                fp.write(
                    os.path.join(root, file),
                    os.path.relpath(
                        os.path.join(root, file),
                        os.path.join(directory, ".."),
                    ),
                )
    return output



"""Tool abstractions."""

import base64
import os
import traceback
import typing as t
from abc import abstractmethod
from pathlib import Path

from pydantic import BaseModel, Field

from composio.tools.base.abs import (
    Action,
    ActionRequest,
    ActionResponse,
    InvalidClassDefinition,
    Tool,
    ToolBuilder,
)
from composio.tools.base.exceptions import ExecutionFailed
from composio.tools.env.host.workspace import Browsers, FileManagers, Shells


class FileModel(BaseModel):
    name: str = Field(
        ...,
        description="File name, contains extension to indetify the file type",
    )
    content: bytes = Field(
        ...,
        description="File content in base64",
    )


class LocalAction(  # pylint: disable=abstract-method
    Action[ActionRequest, ActionResponse],
    abs=True,
):
    """Local action abstraction."""

    _shells: t.Callable[[], Shells]
    _browsers: t.Callable[[], Browsers]
    _filemanagers: t.Callable[[], FileManagers]

    @property
    def shells(self) -> Shells:
        return self._shells()

    @property
    def browsers(self) -> Browsers:
        return self._browsers()

    @property
    def filemanagers(self) -> FileManagers:
        return self._filemanagers()


class LocalToolMeta(type):
    """Tool metaclass."""

    def __init__(  # pylint: disable=self-cls-assignment,unused-argument
        cls,
        name: str,
        bases: t.Tuple,
        dict_: t.Dict,
        autoload: bool = False,
    ) -> None:
        """Initialize action class."""
        if name == "LocalTool":
            return

        ToolBuilder.validate(
            obj=cls,  # type: ignore
            name=name,
            methods=("actions",),
        )
        ToolBuilder.set_metadata(
            obj=cls,  # type: ignore
        )
        ToolBuilder.setup_children(
            obj=cls,  # type: ignore
        )

        if autoload:
            t.cast(t.Type["Tool"], cls).register()  # type: ignore

        if not hasattr(cls, "logo"):
            raise InvalidClassDefinition(f"Please provide logo URL for {name}")

        if "local" not in cls.tags:  # type: ignore
            cls.tags.append("local")  # type: ignore


class LocalToolMixin(Tool):
    @classmethod
    @abstractmethod
    def actions(cls) -> t.List[t.Type[t.Any]]:
        """Get collection of actions for the tool."""

    @classmethod
    def _check_file_uploadable(cls, param: str, model: BaseModel) -> bool:
        return (
            model.model_json_schema()
            .get("properties", {})
            .get(param, {})
            .get("allOf", [{}])[0]
            .get("properties", {})
            or model.model_json_schema()
            .get("properties", {})
            .get(param, {})
            .get("properties", {})
        ) == FileModel.model_json_schema().get("properties")

    @classmethod
    def _process_request(cls, request: t.Dict, model: BaseModel) -> t.Dict:
        """Pre-process request for execution."""
        modified_request_data: t.Dict[str, t.Union[str, t.Dict[str, str]]] = {}
        for param, value in request.items():
            annotations = t.cast(t.Dict, model.model_fields[param].json_schema_extra)
            file_readable = (annotations or {}).get("file_readable", False)
            if file_readable and isinstance(value, str) and os.path.isfile(value):
                _content = Path(value).read_bytes()
                try:
                    _decoded = _content.decode("utf-8")
                except UnicodeDecodeError:
                    _decoded = base64.b64encode(_content).decode("utf-8")
                modified_request_data[param] = _decoded
                continue

            if (
                cls._check_file_uploadable(param=param, model=model)
                and isinstance(value, str)
                and os.path.isfile(value)
            ):
                _content = Path(value).read_bytes()
                modified_request_data[param] = {
                    "name": os.path.basename(value),
                    "content": base64.b64encode(_content).decode("utf-8"),
                }
                continue
            modified_request_data[param] = value
        return modified_request_data

    def execute(
        self,
        action: str,
        params: t.Dict,
        metadata: t.Optional[t.Dict] = None,
    ) -> t.Dict:
        """
        Execute the given action

        :param name: Name of the action.
        :param params: Execution parameters.
        :param metadata: A dictionary containing metadata for action.
        """
        actcls = self._actions.get(action)
        if actcls is None:
            raise ValueError(f"No action found with name `{action}`")

        try:
            metadata = metadata or {}
            instance = actcls(**metadata.get("kwargs", {}))
            if isinstance(instance, LocalAction):
                setattr(instance, "_shells", metadata["_shells"])
                setattr(instance, "_browsers", metadata["_browsers"])
                setattr(instance, "_filemanagers", metadata["_filemanagers"])

            response = instance.execute(
                request=actcls.request.parse(  # type: ignore
                    request=self._process_request(
                        request=params,
                        model=actcls.request.model,  # type: ignore
                    )
                ),
                metadata=metadata,
            )
            return {
                "data": response.model_dump(),
                "error": None,
                "successful": True,
            }
        except ExecutionFailed as e:
            self.logger.error(f"Error executing `{action}`: {e}")
            return {
                "data": None,
                "error": e.message,
                "successful": False,
                **e.extra,
            }
        except Exception as e:
            self.logger.error(f"Error executing `{action}`: {e}")
            self.logger.error(traceback.format_exc())
            return {
                "data": None,
                "error": str(e),
                "successful": False,
            }


class LocalTool(LocalToolMixin, metaclass=LocalToolMeta):
    """Local tool class."""

    gid = "local"
    """Group ID for this tool."""

    tags: t.List[str] = ["local"]
    """Tags for this app."""

    @classmethod
    @abstractmethod
    def actions(cls) -> t.List[t.Type[LocalAction]]:
        """Get collection of actions for the tool."""



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class DeleteChecklistItemRequest(BaseModel):
    """Request schema for `DeleteChecklistItem`"""

    checklist_id: str = Field(
        ...,
        alias="checklist_id",
        description="b8a8-48d8-a0c6-b4200788a683 (uuid)",
    )
    checklist_item_id: str = Field(
        ...,
        alias="checklist_item_id",
        description="e491-47f5-9fd8-d1dc4cedcc6f (uuid)",
    )


class DeleteChecklistItemResponse(BaseModel):
    """Response schema for `DeleteChecklistItem`"""

    data: t.Dict[str, t.Any]


class DeleteChecklistItem(OpenAPIAction):
    """Delete a line item from a task checklist."""

    _tags = ["Task Checklists"]
    _display_name = "delete_checklist_item"
    _request_schema = DeleteChecklistItemRequest
    _response_schema = DeleteChecklistItemResponse

    url = "https://api.clickup.com/api/v2"
    path = "/checklist/{checklist_id}/checklist_item/{checklist_item_id}"
    method = "delete"
    operation_id = "TaskChecklists_removeChecklistItem"
    action_identifier = (
        "/checklist/{checklist_id}/checklist_item/{checklist_item_id}_delete"
    )

    path_params = {
        "checklist_id": "checklist_id",
        "checklist_item_id": "checklist_item_id",
    }
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}
